import os
import random
from datetime import datetime

import pendulum
import pytest
from dotenv import find_dotenv, load_dotenv
from fhir.resources.codeableconcept import CodeableConcept
from fhir.resources.coding import Coding
from fhir.resources.patient import Patient
from fhir.resources.reference import Reference
from pydantic import ValidationError

from fhir_kindling import FhirServer
from fhir_kindling.generators.dataset import DatasetGenerator
from fhir_kindling.generators.field_generator import FieldGenerator
from fhir_kindling.generators.patient import PatientGenerator
from fhir_kindling.generators.resource_generator import (
    FieldValue,
    GeneratorParameters,
    ResourceGenerator,
)
from fhir_kindling.generators.time_series_generator import TimeSeriesGenerator


@pytest.fixture
def api_url():
    """
    Base api url and env vars
    """
    load_dotenv(find_dotenv())

    return os.getenv("FHIR_API_URL", "http://test.fhir.org/r4")


@pytest.fixture
def server(api_url):
    server = FhirServer(
        api_address=api_url,
        client_id=os.getenv("CLIENT_ID"),
        client_secret=os.getenv("CLIENT_SECRET"),
        oidc_provider_url=os.getenv("OIDC_PROVIDER_URL"),
    )
    return server


@pytest.fixture
def covid_code():
    covid_code = CodeableConcept(
        coding=[
            Coding(
                system="http://id.who.int/icd/release/11/mms",
                code="RA01.0",
                display="COVID-19, virus identified",
            )
        ],
        text="COVID-19",
    )
    return covid_code


@pytest.fixture
def vaccination_code():
    covid_code = CodeableConcept(
        coding=[
            Coding(
                system="http://id.who.int/icd/release/11/mms",
                code="XM0GQ8",
                display="COVID-19 vaccine, RNA based",
            )
        ],
        text="COVID vaccination",
    )
    return covid_code


@pytest.fixture
def covid_params(covid_code):
    # generate patients with ids and turn into an iterator to call as a reference
    patients = PatientGenerator(n=100, generate_ids=True).generate()
    patient_ids = iter([patient.id for patient in patients])
    params = GeneratorParameters(
        count=100,
        field_values=[
            FieldValue(field="code", value=covid_code),
        ],
        field_generators=[
            FieldGenerator(
                field="subject",
                generator_function=lambda: {
                    "reference": f"Patient/{next(patient_ids)}"
                },
            )
        ],
    )

    return params, patients, patient_ids


def test_patient_generator():
    patient_generator = PatientGenerator(n=100)
    patients = patient_generator.generate()
    assert len(patients) == 100
    assert isinstance(patients[0], Patient)
    print(patients[0].id)

    generator = PatientGenerator(n=10, age_range=(18, 60))
    patients = generator.generate()

    with pytest.raises(ValueError):
        generator = PatientGenerator(n=10, age_range=("hello", "world"))
        generator.generate()

    assert len(patients) == 10

    organization_reference = Reference(
        reference="Organization/1",
    )

    generator = PatientGenerator(
        n=10, age_range=(18, 60), organisation=organization_reference
    )

    print(generator)
    patients = generator.generate()
    assert len(patients) == 10


def test_generator_field_parameters():
    params = FieldGenerator(
        field="birthdate",
        choices=["2018-01-01", "2018-01-02"],
        choice_probabilities=[0.5, 0.5],
    )

    assert params.field == "birthdate"

    with pytest.raises(ValidationError):
        FieldGenerator(
            field="birthdate",
            choices=["2018-01-01", "2018-01-02"],
            choice_probabilities=[0.5, 0.3, 0.2],
        )

    with pytest.raises(ValidationError):
        FieldGenerator(
            field="birthdate",
            choices=["2018-01-01", "2018-01-02"],
            choice_probabilities=[0.5, 0.6],
        )

    with pytest.raises(ValidationError):
        FieldGenerator(
            field="birthdate",
            choices=["2018-01-01", "2018-01-02"],
            choice_probabilities=[0.5, 0.5],
            generator_function=lambda x: "2018-01-01",
        )

    with pytest.raises(ValidationError):
        FieldGenerator(field="birthdate")


def test_field_generator():
    choices = ["2018-01-01", "2018-01-02"]
    params = FieldGenerator(
        field="birthdate", choices=choices, choice_probabilities=[0.5, 0.5]
    )
    value = params.generate()
    assert value in choices

    params = FieldGenerator(field="birthdate", choices=choices)
    value = params.generate()
    assert value in choices

    params = FieldGenerator(
        field="birthdate", choices=choices, choice_probabilities=[1.0, 0.0]
    )
    for i in range(100):
        value = params.generate()
        assert value == choices[0]

    params = FieldGenerator(field="birthdate", generator_function=lambda: "2018-01-01")
    value = params.generate()
    assert value == "2018-01-01"


def test_generator_with_parameters(covid_params):
    params, patients, patient_ids = covid_params

    generator = ResourceGenerator("Condition", generator_parameters=params)

    resources = generator.generate()

    assert len(resources) == 100


def test_resource_generator(covid_code):
    patient_generator = PatientGenerator(n=100, generate_ids=True)
    patients, references = patient_generator.generate(references=True)

    references_iter = iter(references)

    params = GeneratorParameters(
        count=100,
        field_values=[],
        field_generators=[
            # user patient reference iterator to generate subject field
            FieldGenerator(
                field="subject", generator_function=lambda: next(references_iter)
            ),
        ],
    )

    generator = ResourceGenerator("Condition", generator_parameters=params)
    data = generator.generate()

    assert len(data) == 100

    params = GeneratorParameters(
        count=100,
        field_values=[
            FieldValue(field="subject", value=references),
        ],
    )

    generator = ResourceGenerator("Condition", generator_parameters=params)
    data = generator.generate()
    assert len(data) == 100

    # too few resources in field value
    with pytest.raises(ValueError):
        params = GeneratorParameters(
            count=200,
            field_values=[
                FieldValue(field="subject", value=references),
            ],
        )

        generator = ResourceGenerator("Condition", generator_parameters=params)
        data = generator.generate()

    # duplicate field values should raise error
    with pytest.raises(ValueError):
        patient_2, reference_2 = patient_generator.generate(references=True)
        params = GeneratorParameters(
            count=100,
            field_values=[
                FieldValue(field="subject", value=references),
                FieldValue(field="subject", value=reference_2),
            ],
        )

        generator = ResourceGenerator("Condition", generator_parameters=params)
        data = generator.generate()

    references_iter_2 = iter(references)

    # generator and value targeting the same field should raise error
    with pytest.raises(ValueError):
        patient_2, reference_2 = patient_generator.generate(references=True)
        params = GeneratorParameters(
            count=100,
            field_values=[
                FieldValue(field="subject", value=references),
                FieldGenerator(
                    field="subject", generator_function=lambda: next(references_iter_2)
                ),
            ],
        )

        generator = ResourceGenerator("Condition", generator_parameters=params)
        data = generator.generate()


def test_generate_covid_dataset(vaccination_code, covid_code, server):
    count = 100
    dataset_generator = DatasetGenerator("Patient", n=count)

    covid_params = GeneratorParameters(
        field_values=[
            FieldValue(field="code", value=covid_code),
            # FieldValue(field="subject", value={"reference": "Patient/123"})
        ]
    )

    covid_generator = ResourceGenerator("Condition", generator_parameters=covid_params)
    # add covid conditions to patients
    dataset_generator.add_resource_generator(
        covid_generator, name="covid", depends_on="base", reference_field="subject"
    )

    patients, patient_ids = PatientGenerator(n=count, generate_ids=True).generate(
        references=True
    )

    vaccination_date_generator = FieldGenerator(
        field="occurrenceDateTime",
        generator_function=lambda: pendulum.now().to_date_string(),
    )

    first_vax_params = GeneratorParameters(
        field_values=[
            FieldValue(field="vaccineCode", value=vaccination_code),
            FieldValue(field="status", value="completed"),
        ],
        field_generators=[vaccination_date_generator],
    )
    vaccination_generator = ResourceGenerator(
        "Immunization", generator_parameters=first_vax_params
    )
    print(vaccination_generator)
    dataset_generator.add_resource_generator(
        vaccination_generator,
        name="first_vaccination",
        likelihood=0.8,
        depends_on="base",
        reference_field="patient",
    )

    dataset = dataset_generator.generate()
    print(dataset)
    # pprint(result.dict(exclude_none=True))

    dataset.upload(server)


def test_time_series_generator():
    body_temp_code = CodeableConcept(
        coding=[
            Coding(
                system="http://loinc.org",
                code="99836-5",
                display="Body temperature",
            )
        ]
    )

    body_temp_params = GeneratorParameters(
        field_values=[
            FieldValue(field="code", value=body_temp_code),
            FieldValue(field="status", value="final"),
        ],
        field_generators=[
            FieldGenerator(
                field="valueQuantity",
                generator_function=lambda: {
                    "value": random.randint(36, 40) + random.random(),
                    "unit": "C",
                    "system": "http://unitsofmeasure.org",
                },
            ),
        ],
    )

    body_temp_generator = ResourceGenerator(
        "Observation", generator_parameters=body_temp_params
    )

    body_temp_ts_generator = TimeSeriesGenerator(
        resource_generator=body_temp_generator,
        time_field="effectiveDateTime",
        start=datetime(2021, 1, 1),
        freq="daily",
        n=20,
    )

    result = body_temp_ts_generator.generate()
    print("Result generate n \n\n")
    print(result)
    assert len(result) == 20

    body_temp_range_generator = TimeSeriesGenerator(
        resource_generator=body_temp_generator,
        time_field="effectiveDateTime",
        start=datetime(2021, 1, 1),
        end=datetime(2021, 3, 1),
        freq="daily",
    )
    result = body_temp_range_generator.generate()

    print("Result generate range \n\n")
    print(result)



import random
from datetime import datetime
from time import sleep
from typing import Iterable, Mapping, Union

import httpx
import pendulum


class RetryTransport(httpx.AsyncBaseTransport, httpx.BaseTransport):
    RETRYABLE_METHODS = frozenset(["HEAD", "GET", "PUT", "DELETE", "OPTIONS", "TRACE"])
    RETRYABLE_STATUS_CODES = frozenset([413, 429, 503, 504])

    MAX_BACKOFF_WAIT = 60

    def __init__(
        self,
        wrapped_transport: Union[httpx.BaseTransport, httpx.AsyncBaseTransport],
        max_attempts: int = 10,
        max_backoff_wait: float = MAX_BACKOFF_WAIT,
        backoff_factor: float = 0.1,
        jitter_ratio: float = 0.1,
        respect_retry_after_header: bool = True,
        retryable_methods: Iterable[str] = None,
        retry_status_codes: Iterable[int] = None,
    ) -> None:
        """
        A transport that retries requests that fail with retryable status codes.

        Args:
            wrapped_transport: The transport to wrap.
            max_attempts: The maximum number of attempts to make.
            max_backoff_wait: The maximum amount of time to wait between retries.
            backoff_factor: The amount of time to wait between retries, multiplied by
                the number of attempts made.
            jitter_ratio: The amount of jitter to add to the backoff time. This is
                multiplied by the backoff time and added or subtracted from the backoff
            time.
            respect_retry_after_header: Whether to respect the Retry-After header
                when retrying requests.
            retryable_methods: The HTTP methods that should be retried.
            retry_status_codes: The HTTP status codes that should be retried.
        """
        self.wrapped_transport = wrapped_transport
        if jitter_ratio < 0 or jitter_ratio > 0.5:
            raise ValueError(
                f"jitter ratio should be between 0 and 0.5, actual {jitter_ratio}"
            )

        self.max_attempts = max_attempts
        self.backoff_factor = backoff_factor
        self.respect_retry_after_header = respect_retry_after_header
        self.retryable_methods = (
            frozenset(retryable_methods)
            if retryable_methods
            else self.RETRYABLE_METHODS
        )
        self.retry_status_codes = (
            frozenset(retry_status_codes)
            if retry_status_codes
            else self.RETRYABLE_STATUS_CODES
        )
        self.jitter_ratio = jitter_ratio
        self.max_backoff_wait = max_backoff_wait

    def _calculate_sleep(
        self, attempts_made: int, headers: Union[httpx.Headers, Mapping[str, str]]
    ) -> float:
        retry_after_header = (headers.get("Retry-After") or "").strip()
        if self.respect_retry_after_header and retry_after_header:
            if retry_after_header.isdigit():
                return float(retry_after_header)

            try:
                # convert to local time with pendulum
                parsed_date = pendulum.parse(
                    retry_after_header
                ).astimezone()  # converts to local time# converts to local time
                diff = (parsed_date - datetime.now().astimezone()).total_seconds()
                if diff > 0:
                    return min(diff, self.max_backoff_wait)
            except ValueError:
                pass

        backoff = self.backoff_factor * (2 ** (attempts_made - 1))
        jitter = (backoff * self.jitter_ratio) * random.choice([1, -1])
        total_backoff = backoff + jitter
        return min(total_backoff, self.max_backoff_wait)

    def handle_request(self, request: httpx.Request) -> httpx.Response:
        response = self.wrapped_transport.handle_request(request)

        if request.method not in self.retryable_methods:
            return response

        remaining_attempts = self.max_attempts - 1
        attempts_made = 1

        while True:
            if (
                remaining_attempts < 1
                or response.status_code not in self.retry_status_codes
            ):
                return response

            response.close()

            sleep_for = self._calculate_sleep(attempts_made, response.headers)
            sleep(sleep_for)

            response = self.wrapped_transport.handle_request(request)

            attempts_made += 1
            remaining_attempts -= 1

    async def handle_async_request(self, request: httpx.Request) -> httpx.Response:
        response = await self.wrapped_transport.handle_async_request(request)

        if request.method not in self.retryable_methods:
            return response

        remaining_attempts = self.max_attempts - 1
        attempts_made = 1

        while True:
            if (
                remaining_attempts < 1
                or response.status_code not in self.retry_status_codes
            ):
                return response

            response.close()

            sleep_for = self._calculate_sleep(attempts_made, response.headers)
            sleep(sleep_for)

            response = await self.wrapped_transport.handle_async_request(request)

            attempts_made += 1
            remaining_attempts -= 1



import random

import pendulum

from fhir_kindling.benchmark.constants import Codes
from fhir_kindling.generators.dataset import DatasetGenerator
from fhir_kindling.generators.field_generator import FieldGenerator
from fhir_kindling.generators.resource_generator import (
    FieldValue,
    GeneratorParameters,
    ResourceGenerator,
)
from fhir_kindling.generators.time_series_generator import TimeSeriesGenerator

N_BASE_RESOURCES = 100


def generate_benchmark_data(n_patients: int = N_BASE_RESOURCES) -> DatasetGenerator:
    dataset_generator = DatasetGenerator("Patient", n=n_patients)

    # covid
    covid_params = GeneratorParameters(
        field_values=[
            FieldValue(field="code", value=Codes.COVID.value),
        ]
    )
    covid_generator = ResourceGenerator("Condition", generator_parameters=covid_params)
    dataset_generator.add_resource_generator(
        covid_generator, name="covid", depends_on="base", reference_field="subject"
    )

    # covid vaccination(s)

    vaccination_date_generator = FieldGenerator(
        field="occurrenceDateTime",
        generator_function=lambda: pendulum.now().to_date_string(),
    )

    # first shot covid vaccine

    first_vax_params = GeneratorParameters(
        field_values=[
            FieldValue(field="vaccineCode", value=Codes.COVID_VACC_RNA.value),
            FieldValue(field="status", value="completed"),
        ],
        field_generators=[vaccination_date_generator],
    )
    vaccination_generator = ResourceGenerator(
        "Immunization",
        generator_parameters=first_vax_params,
    )

    dataset_generator.add_resource_generator(
        vaccination_generator,
        "vacc-mrna-1",
        depends_on="base",
        likelihood=0.7,
        reference_field="patient",
    )

    # second shot
    second_vax_params = GeneratorParameters(
        field_values=[
            FieldValue(field="vaccineCode", value=Codes.COVID_VACC_RNA.value),
            FieldValue(field="status", value="completed"),
        ],
        field_generators=[vaccination_date_generator],
    )
    second_vaccination_generator = ResourceGenerator(
        "Immunization", generator_parameters=second_vax_params
    )

    dataset_generator.add_resource_generator(
        second_vaccination_generator,
        "vacc-mrna-2",
        depends_on=["base", "vacc-mrna-1"],
        likelihood=0.9,
        reference_field=["patient", None],
    )

    # third shot
    second_vax_params = GeneratorParameters(
        field_values=[
            FieldValue(field="vaccineCode", value=Codes.COVID_VACC_RNA.value),
            FieldValue(field="status", value="completed"),
        ],
        field_generators=[vaccination_date_generator],
    )
    third_vaccination_generator = ResourceGenerator(
        "Immunization", generator_parameters=second_vax_params
    )

    dataset_generator.add_resource_generator(
        third_vaccination_generator,
        "vacc-mrna-3",
        depends_on=["base", "vacc-mrna-1", "vacc-mrna-2"],
        reference_field=["patient", None, None],
        likelihood=0.7,
    )
    # generate encounters
    emergency_encounter_period_generator = FieldGenerator(
        field="period",
        generator_function=lambda: {
            "start": pendulum.now().subtract(days=730).to_date_string(),
            "end": pendulum.now().subtract(days=729).to_date_string(),
        },
    )

    # emergency encounter
    emergency_encounter_params = GeneratorParameters(
        field_values=[
            FieldValue(field="class", value=Codes.EMERGENCY_ENCOUNTER.value),
            FieldValue(field="status", value="finished"),
        ],
        field_generators=[emergency_encounter_period_generator],
    )

    emergency_encounter_generator = ResourceGenerator(
        "Encounter", generator_parameters=emergency_encounter_params
    )

    dataset_generator.add_resource_generator(
        emergency_encounter_generator,
        "emergency-encounter",
        depends_on="base",
        reference_field="subject",
    )

    icu_encounter_period_generator = FieldGenerator(
        field="period",
        generator_function=lambda: {
            "start": pendulum.now().subtract(days=720).to_date_string(),
            "end": pendulum.now().subtract(days=710).to_date_string(),
        },
    )

    # icu encounter
    icu_encounter_params = GeneratorParameters(
        field_values=[
            FieldValue(field="class", value=Codes.ICU_ENCOUNTER.value),
            FieldValue(
                field="type", value=Codes.ICU_ENCOUNTER_TYPE.value, list_field=True
            ),
            FieldValue(field="status", value="finished"),
        ],
        field_generators=[icu_encounter_period_generator],
    )

    icu_encounter_generator = ResourceGenerator(
        "Encounter", generator_parameters=icu_encounter_params
    )

    dataset_generator.add_resource_generator(
        icu_encounter_generator,
        "icu-encounter",
        depends_on=["base", "emergency-encounter"],
        reference_field=["subject", None],
    )

    # generate observations

    # blood oxygen saturation
    blood_oxygen_params = GeneratorParameters(
        field_values=[
            FieldValue(field="code", value=Codes.OXYGEN_SATURATION.value),
            FieldValue(field="status", value="final"),
        ],
        field_generators=[
            FieldGenerator(
                field="valueQuantity",
                generator_function=lambda: {
                    "value": random.randint(90, 100),
                    "unit": "%",
                },
            ),
        ],
    )
    blood_oxygen_saturation_generator = ResourceGenerator(
        "Observation", generator_parameters=blood_oxygen_params
    )

    bo_time_series_generator = TimeSeriesGenerator(
        resource_generator=blood_oxygen_saturation_generator,
        start=pendulum.now().subtract(days=720),
        n=10,
        time_field="effectiveDateTime",
        freq="daily",
    )

    dataset_generator.add_resource_generator(
        bo_time_series_generator,
        "blood-oxygen-saturation",
        depends_on="base",
        reference_field="subject",
    )

    # body temperature
    body_temperature_params = GeneratorParameters(
        field_values=[
            FieldValue(field="code", value=Codes.BODY_TEMPERATURE.value),
            FieldValue(field="status", value="final"),
        ],
        field_generators=[
            FieldGenerator(
                field="valueQuantity",
                generator_function=lambda: {
                    "value": random.randint(36, 40) + random.random(),
                    "unit": "°C",
                },
            ),
        ],
    )

    body_temperature_generator = ResourceGenerator(
        "Observation", generator_parameters=body_temperature_params
    )

    dataset_generator.add_resource_generator(
        body_temperature_generator,
        "body-temperature",
        depends_on="icu-encounter",
        reference_field="encounter",
    )

    # respiratory rate
    respiratory_rate_params = GeneratorParameters(
        field_values=[
            FieldValue(field="code", value=Codes.RESPIRATORY_RATE.value),
            FieldValue(field="status", value="final"),
        ],
        field_generators=[
            FieldGenerator(
                field="valueQuantity",
                generator_function=lambda: {
                    "value": random.randint(12, 30),
                    "unit": "breaths/min",
                },
            ),
        ],
    )

    respiratory_rate_generator = ResourceGenerator(
        "Observation", generator_parameters=respiratory_rate_params
    )

    dataset_generator.add_resource_generator(
        respiratory_rate_generator,
        "respiratory-rate",
        depends_on="icu-encounter",
        reference_field="encounter",
    )

    # heart rate
    heart_rate_params = GeneratorParameters(
        field_values=[
            FieldValue(field="code", value=Codes.HEART_RATE.value),
            FieldValue(field="status", value="final"),
        ],
        field_generators=[
            FieldGenerator(
                field="valueQuantity",
                generator_function=lambda: {
                    "value": random.randint(60, 100),
                    "unit": "beats/min",
                },
            ),
        ],
    )

    heart_rate_generator = ResourceGenerator(
        "Observation", generator_parameters=heart_rate_params
    )

    dataset_generator.add_resource_generator(
        heart_rate_generator,
        name="heart-rate",
        depends_on="icu-encounter",
        reference_field="encounter",
    )

    return dataset_generator



from .query_async import FhirQueryAsync
from .query_parameters import (
    FhirQueryParameters,
    FieldParameter,
    IncludeParameter,
    QueryOperators,
    ReverseChainParameter,
)
from .query_response import QueryResponse
from .query_sync import FhirQuerySync



from enum import Enum

from fhir.resources.codeableconcept import CodeableConcept
from fhir.resources.coding import Coding


class BenchmarkOperations(str, Enum):
    GENERATE = "generate"
    INSERT = "insert"
    BATCH_INSERT = "batch_insert"
    DATASET_INSERT = "dataset_insert"
    QUERY = "query"
    UPDATE = "update"
    DELETE = "delete"
    BATCH_DELETE = "batch_delete"


class DefaultQueries(str, Enum):
    PATIENTS = "Patient?"
    # Filter patients by birthdate
    PATIENT_BY_BIRTHDATE = "Patient?birthdate=lt2020&birthdate=gt1970"
    # Include observaations
    PATIENTS_WITH_OBSERVATIONS = "Patient?_revinclude=Observation:subject"
    # Patients with emergency encounter with immunization included
    PATIENTS_WITH_EMERGENCY_ENCOUNTER = (
        "Patient?_has:Encounter:subject:class=EMER&_revinclude=Immunization:patient"
    )
    # Immunization for patients with covid code
    IMMUNIZATION_FOR_COVID_PATIENTS = (
        "Patient?_has:Condition:subject:code=RA01.0&_revinclude=Immunization:patient"
    )
    # get observations with am observed breating rate of 20 or less and rev include patients
    PATIENTS_BY_OBSERVATION_VALUE = (
        "Observation?component-code-value-quantity="
        "http://loinc.org|9279-1$lt20&_include=Observation:subject"
    )  # noqa


class Codes(Enum):
    COVID = CodeableConcept(
        coding=[
            Coding(
                system="http://id.who.int/icd/release/11/mms",
                code="RA01.0",
                display="COVID-19, virus identified",
            )
        ],
        text="COVID-19",
    )
    COVID_VACC_RNA = CodeableConcept(
        coding=[
            Coding(
                system="http://id.who.int/icd/release/11/mms",
                code="XM0GQ8",
                display="COVID-19 vaccine, RNA based",
            )
        ],
        text="COVID vaccination",
    )
    EMERGENCY_ENCOUNTER = Coding(
        system="http://terminology.hl7.org/ValueSet/v3-ActEncounterCode",
        code="EMER",
        display="emergency",
    )

    ICU_ENCOUNTER = Coding(
        system="http://terminology.hl7.org/ValueSet/v3-ActEncounterCode",
        code="ACUTE",
        display="Acute inpatient encounter",
    )
    ICU_ENCOUNTER_TYPE = [
        CodeableConcept(
            coding=[
                Coding(
                    system="http://loinc.org",
                    code="99222",
                    display="Inpatient Hospitalization",
                )
            ],
        )
    ]

    OXYGEN_SATURATION = CodeableConcept(
        coding=[
            Coding(
                system="http://loinc.org",
                code="55284-4",
                display="Oxygen saturation in Arterial blood by Pulse oximetry",
            )
        ],
    )

    BODY_TEMPERATURE = CodeableConcept(
        coding=[
            Coding(
                system="http://loinc.org",
                code="99836-5",
                display="Body temperature",
            )
        ]
    )

    RESPIRATORY_RATE = CodeableConcept(
        coding=[
            Coding(
                system="http://loinc.org",
                code="9279-1",
                display="Respiratory rate",
            )
        ]
    )

    HEART_RATE = CodeableConcept(
        coding=[
            Coding(
                system="http://loinc.org",
                code="8802-8",
                display="Heart rate",
            )
        ],
    )



import os

from fhir_kindling import FhirServer
from fhir_kindling.fhir_server.transfer import reference_graph, transfer


def test_reference_graph(fhir_server):
    """
    Test the reference graph
    """
    # get all conditions
    conditions = fhir_server.query("Condition").all().resource_list

    # get the reference graph
    graph = reference_graph(conditions)
    print("Nodes:")
    print(graph.nodes)

    print("Edges:")
    print(graph.edges)

    print("Graph:")
    print(graph.graph)

    assert len(graph.nodes) == len(conditions) * 2


def test_transfer_resources(fhir_server):
    """
    Test the transfer resources
    """
    # get all conditions
    conditions = fhir_server.query("Condition").all().resource_list

    target = FhirServer(api_address=os.getenv("TRANSFER_SERVER_URL"))

    response = transfer(source=fhir_server, target=target, resources=conditions)

    print(response)
    print(response.linkage)
    # save linkages
    response.save_linkage("linkage.json")

    os.remove("linkage.json")



from typing import TYPE_CHECKING

import plotly.graph_objects as go
from plotly.subplots import make_subplots

if TYPE_CHECKING:
    from fhir_kindling.benchmark.results import BenchmarkResults


def plot_benchmark_results(results: "BenchmarkResults") -> go.Figure:
    """Create a multi-panel plotly figure that depicts the results of the benchmark. The figure displays
    the timings for different crud & query operations and compares the timings across the different servers.

    Args:
        results: benchmark results
    """
    fig = make_subplots(
        rows=5,
        cols=1,
        subplot_titles=(
            "Single Insert",
            "Batch Insert",
            "Dataset Insert" "Update Single",
            "Search",
        ),
    )

    add_single_insert_traces(fig, results)
    add_batch_insert_traces(fig, results)
    add_dataset_insert_traces(fig, results)
    add_query_traces(fig, results)
    add_delete_traces(fig, results)

    fig.update_layout(
        title_text="FHIR Server Benchmark Results",
        height=1200,
        width=1000,
        showlegend=True,
        legend_tracegroupgap=190,
    )
    return fig


def add_single_insert_traces(fig: go.Figure, results: "BenchmarkResults"):
    """Create traces for each server detailing the timings for inserting a single resource

    Args:
        results: the results in which the data is stored

    Returns:
        List of plotly bar charts that depict the timings for each server
    """
    for server, result in results.insert_single.items():
        tr = go.Bar(y=result, name=server, legendgroup="1")
        fig.add_trace(tr, row=1, col=1)


def add_batch_insert_traces(fig: go.Figure, results: "BenchmarkResults"):
    """Create traces for each server detailing the timings for inserting a single resource

    Args:
        results: the results in which the data is stored

    Returns:
        List of plotly bar charts that depict the timings for each server
    """
    for server, result in results.batch_insert.items():
        tr = go.Bar(y=result, name=server, legendgroup="2")
        fig.add_trace(tr, row=2, col=1)


def add_dataset_insert_traces(fig: go.Figure, results: "BenchmarkResults"):
    """Create traces for each server detailing the timings for inserting a single resource

    Args:
        fig: _description_
        results: _description_
    """
    ds_fig = go.Bar(
        y=list(results.dataset_insert.values()),
        x=list(results.dataset_insert.keys()),
        name="Dataset Insert",
    )
    fig.add_trace(ds_fig, row=3, col=1)


def add_query_traces(fig: go.Figure, results: "BenchmarkResults"):
    query_results = results.query
    queries = list(list(query_results.values())[0].keys())

    for server, result in query_results.items():
        # average time for each query
        avg_times = [sum(result[q]) / len(result[q]) for q in queries]

        tr = go.Bar(y=avg_times, x=queries, name=server, legendgroup="4")
        fig.add_trace(tr, row=4, col=1)


def add_delete_traces(fig: go.Figure, results: "BenchmarkResults"):
    delete_results = results.delete
    x = list(delete_results.keys())
    y = list(delete_results.values())
    tr = go.Bar(y=y, x=x, name="Delete", legendgroup="5")
    fig.add_trace(tr, row=5, col=1)



from typing import Any, Callable, List, Union

import fhir.resources
import httpx
import orjson
import xmltodict
from fhir.resources import FHIRAbstractModel
from fhir.resources.fhirresourcemodel import FHIRResourceModel

from fhir_kindling.fhir_query.base import FhirQueryBase
from fhir_kindling.fhir_query.query_parameters import (
    FhirQueryParameters,
)
from fhir_kindling.fhir_query.query_response import (
    OutputFormats,
    QueryResponse,
    ResponseStatusCodes,
)


class FhirQuerySync(FhirQueryBase):
    def __init__(
        self,
        base_url: str,
        resource: Union[
            FHIRResourceModel, fhir.resources.FHIRAbstractModel, str
        ] = None,
        query_parameters: FhirQueryParameters = None,
        auth: httpx.Auth = None,
        headers: dict = None,
        client: httpx.Client = None,
        output_format: str = "json",
        proxies: Union[str, dict] = None,
    ):
        """Initialize an sync FHIR query object.

        Args:
            base_url: Base URL of the FHIR server to query.
            resource: Base resource to build the query on. Defaults to None.
            query_parameters: Query parameters object that fully describes a FHIR query. Defaults to None.
            auth: httpx auth object to authenticate the requests. Defaults to None.
            headers: Optional additional headers to add to the request. Defaults to None.
            output_format: Response format of the query. Defaults to "json".
            client: httpx Client passed from the server. Defaults to None.
            proxies: List of proxies to use. Defaults to None.
        """

        super().__init__(
            base_url,
            resource,
            query_parameters,
            auth,
            headers,
            output_format,
        )
        self.proxies = proxies
        self.client = None
        if client:
            self.client = client
        else:
            self.client = self._setup_client()

    def all(
        self,
        page_callback: Union[
            Callable[[List[FHIRAbstractModel]], Any], Callable[[], Any], None
        ] = None,
        count: int = None,
    ) -> QueryResponse:
        """
        Execute the query and return all results matching the query parameters.

        Args:
            page_callback: if this argument is set the given callback function will be called for each page of results
            count: number of results in a page, default value of 50 is used when page_callback is set but no count is
        Returns:
            QueryResponse object containing all resources matching the query, as well os optional included
            resources.

        """
        self._limit = None
        self._count = count
        return self._execute_query(page_callback=page_callback, count=count)

    def limit(
        self,
        n: int,
        page_callback: Union[
            Callable[[List[FHIRAbstractModel]], Any], Callable[[], Any], None
        ] = None,
        count: int = None,
    ) -> QueryResponse:
        """
        Execute the query and return the first n results matching the query parameters.
        Args:
            n: number of resources to return
            page_callback: if this argument is set the given callback function will be called for each page of results
            count: number of results in a page, default value of 50 is used when page_callback is set but no count is

        Returns:
            QueryResponse object containing the first n resources matching the query, as well os optional included
            resources.

        """
        self._limit = n
        self._count = count
        return self._execute_query(page_callback=page_callback, count=count)

    def first(self) -> QueryResponse:
        """
        Return the first resource matching the query parameters.
        Returns:
            QueryResponse object containing the first resource matching the query

        """
        self._limit = 1
        return self._execute_query()

    def count(self) -> int:
        self._count = 0
        response = self.client.get(self._make_query_string() + "&_summary=count")
        response.raise_for_status()
        return response.json()["total"]

    def _setup_client(self):
        if self.client:
            return self.client
        headers = self.headers if self.headers else {}
        headers["Content-Type"] = "application/fhir+json"
        client = httpx.Client(
            auth=self.auth, headers=headers, proxies=self.proxies, timeout=None
        )
        return client

    def _execute_query(
        self,
        page_callback: Union[
            Callable[[List[FHIRAbstractModel]], Any], Callable[[], Any], None
        ] = None,
        count: int = None,
    ) -> QueryResponse:
        r = self.client.get(self.query_url)
        r.raise_for_status()

        response = self._resolve_response_pagination(r, page_callback, count)
        return response

    def _resolve_response_pagination(
        self,
        initial_response: httpx.Response,
        page_callback: Union[
            Callable[[List[FHIRAbstractModel]], Any], Callable[[], Any], None
        ] = None,
        count: int = None,
    ) -> QueryResponse:
        if self.output_format == OutputFormats.JSON:
            response = self._resolve_json_pagination(
                initial_response, page_callback, count
            )

        else:
            response = self._resolve_xml_pagination(initial_response)

        return QueryResponse(
            response=response,
            query_params=self.query_parameters,
            count=count,
            limit=self._limit,
            output_format=self.output_format,
        )

    def _resolve_json_pagination(
        self,
        initial_response: httpx.Response,
        page_callback: Union[
            Callable[[List[FHIRAbstractModel]], Any], Callable[[], Any], None
        ] = None,
        count: int = None,
    ) -> dict:
        response_json = orjson.loads(initial_response.content)
        link = response_json.get("link", None)
        # If there is a link, get the next page otherwise return the response

        if not link:
            self.status_code = ResponseStatusCodes.OK
            return response_json
        else:
            entries = []
            initial_entry = response_json.get("entry", None)
            if not initial_entry:
                self.status_code = ResponseStatusCodes.NOT_FOUND
                return response_json
            else:
                self.status_code = ResponseStatusCodes.OK
                response_entries = response_json["entry"]
                entries.extend(response_entries)
                self._execute_callback(response_entries, page_callback)
            # if the limit is reached, stop resolving the pagination
            if self._limit and len(entries) >= self._limit:
                response_entries = response_json["entry"][: self._limit]
                response_json["entry"] = response_entries
                self._execute_callback(response_entries, page_callback)
                return response_json
            # query the linked page and add the entries to the response

            while response_json.get("link", None):
                if self._limit and len(entries) >= self._limit:
                    break
                next_page = next(
                    (
                        link
                        for link in response_json["link"]
                        if link.get("relation", None) == "next"
                    ),
                    None,
                )
                if next_page:
                    response_json = self.client.get(next_page["url"]).json()
                    response_entries = response_json["entry"]
                    entries.extend(response_entries)
                    self._execute_callback(response_entries, page_callback)
                else:
                    break

            response_json["entry"] = entries[: self._limit] if self._limit else entries
            return response_json

    def _resolve_xml_pagination(self, server_response: httpx.Response) -> str:
        # parse the xml response and extract the initial entries
        initial_response = xmltodict.parse(server_response.text)
        entries = initial_response["Bundle"].get("entry")

        # if there are no entries, return the initial response
        if not entries:
            self.status_code = ResponseStatusCodes.NOT_FOUND
            print(
                f"No resources match the query - query url: {self.query_parameters.to_query_string()}"
            )
            return server_response.text
        else:
            self.status_code = ResponseStatusCodes.OK
        response = initial_response
        # resolve the pagination
        while True:
            next_page = False
            for link in response["Bundle"]["link"]:
                # if there is a next page, get the next page
                if not isinstance(link, dict):
                    break
                relation_dict = link.get("relation", None)

                if relation_dict.get("@value") == "next":
                    # get url and extend with xml format
                    url = link["url"]["@value"]
                    url = url + "&_format=xml"
                    r = self.client.get(url)
                    r.raise_for_status()
                    response = xmltodict.parse(r.text)
                    if not response["Bundle"].get("entry", None):
                        break
                    added_entries = response["Bundle"]["entry"]
                    entries.extend(added_entries)
                    # Stop resolving the pagination when the limit is reached
                    if self._limit:
                        next_page = len(entries) < self._limit
                    else:
                        next_page = True

            if not next_page:
                break
        # added the paginated resources to the initial response
        initial_response["Bundle"]["entry"] = (
            entries[: self._limit] if self._limit else entries
        )
        full_response_xml = xmltodict.unparse(initial_response, pretty=True)
        return full_response_xml



import os
import re
from typing import Iterable, List, Union

import fhir.resources
import httpx
from authlib.integrations.httpx_client import OAuth2Client
from authlib.oauth2.rfc6749 import OAuth2Token
from authlib.oauth2.rfc7523 import ClientSecretJWT
from fhir.resources import FHIRAbstractModel, construct_fhir_element
from fhir.resources.bundle import Bundle, BundleEntry
from fhir.resources.capabilitystatement import CapabilityStatement
from fhir.resources.fhirresourcemodel import FHIRResourceModel
from fhir.resources.reference import Reference
from fhir.resources.resource import Resource
from tqdm import tqdm

from fhir_kindling.fhir_query import FhirQueryAsync, FhirQuerySync
from fhir_kindling.fhir_query.query_parameters import FhirQueryParameters
from fhir_kindling.fhir_server.auth import BearerAuth, auth_info_from_env
from fhir_kindling.fhir_server.server_responses import (
    BundleCreateResponse,
    ResourceCreateResponse,
    TransferResponse,
)
from fhir_kindling.fhir_server.summary import (
    ServerSummary,
    create_server_summary,
    create_server_summary_async,
)
from fhir_kindling.fhir_server.transactions import (
    TransactionMethod,
    TransactionType,
    make_transaction_bundle,
)
from fhir_kindling.fhir_server.transfer import transfer
from fhir_kindling.serde.json import json_dict
from fhir_kindling.util.retry_transport import RetryTransport


class FhirServer:
    def __init__(
        self,
        api_address: str,
        username: Union[str, None] = None,
        password: Union[str, None] = None,
        token: Union[str, None] = None,
        client_id: Union[str, None] = None,
        client_secret: Union[str, None] = None,
        oidc_provider_url: Union[str, None] = None,
        auth: Union[httpx.Auth, None] = None,
        headers: Union[dict, None] = None,
        proxies: Union[dict, str, None] = None,
        timeout: Union[int, None] = None,
        fhir_server_type: str = "hapi",
        retry_status_codes: Union[Iterable[int], None] = None,
        retryable_methods: Union[Iterable[str], None] = None,
        max_atttempts: int = 5,
        max_backoff_wait: float = 60,
        backoff_factor: float = 0.1,
        jitter_ratio: float = 0.1,
        respect_retry_after_header: bool = True,
    ):
        """
        Initialize a FHIR server connection
        Args:
            api_address: the base endpoint of the fhir server REST API
            username: username for basic auth
            password: password for basic auth
            token: token for static token auth
            client_id: client id for oauth2
            client_secret: client secret for oauth2
            oidc_provider_url: provider url for oauth2
            auth: optional auth object to authenticate against a server
            headers: optional additional headers to be added to the session
            proxies: optional proxies to be added to the session
            timeout: optional timeout for the session default is None
            fhir_server_type: type of fhir server (hapi, blaze, etc.)
            retry_status_codes: optional list of status codes to retry on
            max_atttempts: optional number of times to retry
            retry_wait: optional number of seconds to wait between retries
        """

        # server definition values
        self.fhir_server_type = fhir_server_type
        self.api_address = self._validate_api_address(api_address)
        self._meta_data = None

        # possible basic auth class vars
        self.username = username
        self.password = password

        # static token
        self.token = token

        # oauth2 auth vars
        self.client_id = client_id
        self.client_secret = client_secret
        self.oidc_provider_url = oidc_provider_url
        self.oauth_token: OAuth2Token = None

        # retry vars
        self.retry_status_codes = retry_status_codes
        self.retryable_methods = retryable_methods
        self.max_attempts = max_atttempts
        self.max_backoff_wait = max_backoff_wait
        self.backoff_factor = backoff_factor
        self.jitter_ratio = jitter_ratio
        self.respect_retry_after_header = respect_retry_after_header

        self._auth = auth
        self._headers = headers
        self._proxies = proxies
        self._timeout = timeout

    @classmethod
    def from_env(cls, no_auth: bool = False) -> "FhirServer":
        api_address = _api_address_from_env()
        server_type = os.getenv("FHIR_SERVER_TYPE")

        if no_auth:
            return FhirServer(api_address=api_address, fhir_server_type=server_type)
        else:
            env_auth = auth_info_from_env()
            # static token
            if isinstance(env_auth, str):
                return cls(
                    api_address=api_address,
                    token=env_auth,
                    fhir_server_type=server_type,
                )
            # username and password
            elif isinstance(env_auth, tuple) and len(env_auth) == 2:
                return cls(
                    api_address=api_address,
                    username=env_auth[0],
                    password=env_auth[1],
                    fhir_server_type=server_type,
                )
            # oauth2/oidc
            elif isinstance(env_auth, tuple) and len(env_auth) == 3:
                return cls(
                    api_address=api_address,
                    client_id=env_auth[0],
                    client_secret=env_auth[1],
                    oidc_provider_url=env_auth[2],
                    fhir_server_type=server_type,
                )
            else:
                raise EnvironmentError(
                    "Authentication information could not be loaded from environment"
                )

    def query(
        self,
        resource: Union[Resource, FHIRAbstractModel, str] = None,
        query_string: str = None,
        query_parameters: FhirQueryParameters = None,
        output_format: str = "json",
    ) -> FhirQuerySync:
        """
        Initialize a FHIR query against the server with the given resource, query parameters or query string

        Args:
            output_format: the output format to request from the fhir server (json or xml) defaults to json
            query_string: preformatted query string to execute against the servers REST API
            query_parameters: optionally pass in a query parameters object to use for the query
            resource: the FHIR resource to query from the server

        Returns:
            a FhirQuerySync object that can be further modified with filters and conditions before being executed
            against the server
        """

        self._validate_query_input(resource, query_parameters, query_string)
        query_parameters = self._setup_query_parameters(
            resource, query_string, query_parameters
        )

        query = FhirQuerySync(
            base_url=self.api_address,
            auth=self.auth,
            query_parameters=query_parameters,
            output_format=output_format,
            proxies=self._proxies,
            headers=self._headers,
            client=self._sync_client(),
        )

        return query

    def query_async(
        self,
        resource: Union[Resource, FHIRAbstractModel, str] = None,
        query_string: str = None,
        query_parameters: FhirQueryParameters = None,
        output_format: str = "json",
    ) -> FhirQueryAsync:
        """
        Initialize an asynchronous FHIR query against the server with the given resource,
        query parameters or query string

        Args:
            output_format: the output format to request from the fhir server (json or xml) defaults to json
            query_string: preformatted query string to execute against the servers REST API
            query_parameters: optionally pass in a query parameters object to use for the query
            resource: the FHIR resource to query from the server

        Returns:
            a FhirQueryAsync object that can be further modified with filters and conditions before being executed
            against the server
        """

        self._validate_query_input(resource, query_parameters, query_string)
        query_parameters = self._setup_query_parameters(
            resource, query_string, query_parameters
        )

        query = FhirQueryAsync(
            self.api_address,
            auth=self.auth,
            headers=self._headers,
            query_parameters=query_parameters,
            output_format=output_format,
            proxies=self._proxies,
            client=self._async_client(),
        )

        return query

    def _setup_query_parameters(
        self,
        resource: Union[Resource, FHIRAbstractModel, str] = None,
        query_string: str = None,
        query_parameters: FhirQueryParameters = None,
    ) -> FhirQueryParameters:
        """Initialize a FhirQueryParameters object from the given input

        Args:
            resource: Resource to use as base either string or model. Defaults to None.
            query_string: Query string to transform into query parameters. Defaults to None.
            query_parameters: Query parameters object that gets returned directly. Defaults to None.

        Returns:
            _description_
        """
        if resource:
            if isinstance(resource, str):
                query_parameters = FhirQueryParameters(resource=resource)
            elif isinstance(resource, (Resource, FHIRAbstractModel)):
                query_parameters = FhirQueryParameters(resource=resource.resource_type)

        elif query_string:
            query_parameters = FhirQueryParameters.from_query_string(query_string)

        return query_parameters

    def raw_query(
        self, query_string: str, output_format: str = "json"
    ) -> FhirQuerySync:
        """
        Execute a raw query string against the server

        Args:
            query_string:
            output_format:
        Returns:
            a FhirQuerySync object that can be further modified with filters and conditions before being executed
            against the server
        """

        query_parameters = FhirQueryParameters.from_query_string(query_string)
        query = FhirQuerySync(
            base_url=self.api_address,
            query_parameters=query_parameters,
            output_format=output_format,
            auth=self.auth,
            proxies=self._proxies,
            headers=self._headers,
            client=self._sync_client(),
        )
        return query

    def raw_query_async(
        self, query_string: str, output_format: str = "json"
    ) -> FhirQueryAsync:
        """
        Asynchronously Execute a raw query string against the server

        Args:
            query_string: query string defining the query to execute
            output_format: the output format to request from the fhir server (json or xml) defaults to json
        Returns:
            a FhirQueryAsync object that can be further modified with filters and conditions before being executed
            against the server

        """

        query_parameters = FhirQueryParameters.from_query_string(query_string)
        query = FhirQueryAsync(
            base_url=self.api_address,
            resource=query_parameters.resource,
            query_parameters=query_parameters,
            output_format=output_format,
            auth=self.auth,
            proxies=self._proxies,
            client=self._async_client(),
        )
        return query

    def get(self, reference: Union[str, Reference]) -> FHIRAbstractModel:
        """
        Get a resource from the server specified by the given reference {ResourceType}/{id}

        Args:
            reference: reference to the resource, either a Reference object or a string of the form {ResourceType}/{id}

        Returns:
            the resource from the server specified by the reference

        """
        if isinstance(reference, Reference):
            reference = reference.reference
        with self._sync_client() as client:
            r = client.get(f"{self.api_address}/{reference}")
        r.raise_for_status()
        resource_dict = r.json()
        resource = construct_fhir_element(resource_dict["resourceType"], resource_dict)
        return resource

    async def get_async(self, reference: Union[str, Reference]) -> FHIRAbstractModel:
        """
        Asynchronously get a resource from the server specified by the given reference {ResourceType}/{id}

        Args:
            reference: reference to the resource, either a Reference object or a string of the form {ResourceType}/{id}

        Returns:
            the resource from the server specified by the reference

        """
        if isinstance(reference, Reference):
            reference = reference.reference
        async with self._async_client() as client:
            r = await client.get(f"{self.api_address}/{reference}")
            r.raise_for_status()
        resource_dict = r.json()
        resource = construct_fhir_element(resource_dict["resourceType"], resource_dict)
        return resource

    def get_many(
        self, references: List[Union[str, Reference]]
    ) -> List[FHIRAbstractModel]:
        """
        Get a list of resources from the server specified by the given references

        Args:
            references: list of references to the resources, either a Reference object or a string of the form
                `{ResourceType}/{id}`

        Returns:
            list of resources corresponding to the references

        """
        str_references = [
            reference if isinstance(reference, str) else reference.reference
            for reference in references
        ]

        get_many_transaction = make_transaction_bundle(
            method=TransactionMethod.GET,
            transaction_type=TransactionType.BATCH,
            references=str_references,
        )
        with self._sync_client() as client:
            r = client.post(self.api_address, json=json_dict(get_many_transaction))
            r.raise_for_status()
        entries = r.json()["entry"]
        resources = [
            construct_fhir_element(entry["resource"]["resourceType"], entry["resource"])
            for entry in entries
        ]

        return resources

    async def get_many_async(
        self, references: List[Union[str, Reference]]
    ) -> List[FHIRAbstractModel]:
        """
        Asynchronously get a list of resources from the server specified by the given references

        Args:
            references: list of references to the resources, either a Reference object or a string of the form
                `{ResourceType}/{id}`

        Returns:
            list of resources corresponding to the references

        """
        str_references = [
            reference if isinstance(reference, str) else reference.reference
            for reference in references
        ]
        get_many_transaction = make_transaction_bundle(
            method=TransactionMethod.GET,
            transaction_type=TransactionType.BATCH,
            references=str_references,
        )

        async with self._async_client() as client:
            response = await client.post(
                self.api_address, json=get_many_transaction.dict(exclude_none=True)
            )

        # construct the list of resources from the server response
        resources = [
            construct_fhir_element(entry["resource"]["resourceType"], entry["resource"])
            for entry in response.json()["entry"]
        ]
        return resources

    def add(self, resource: Union[Resource, dict]) -> ResourceCreateResponse:
        """
        Upload a resource to the server

        Args:
            resource: dictionary containing the resource or FHIR resource object to be uploaded to the server

        Returns:
            CreateResponse containing the server response and the resource that was uploaded

        """
        # parse the resource given as dictionary into a fhir resource model
        if isinstance(resource, dict):
            resource_type = resource.get("resourceType")
            if not resource_type:
                raise ValueError("No resource type defined in resource dictionary")
            resource = fhir.resources.construct_fhir_element(resource_type, resource)
        else:
            resource = resource.validate(resource)
        response = self._upload_resource(resource)
        response.raise_for_status()

        return ResourceCreateResponse(
            server_response_dict=dict(response.headers), resource=resource
        )

    async def add_async(
        self, resource: Union[Resource, dict]
    ) -> ResourceCreateResponse:
        """
        Asynchronously upload a resource to the server

        Args:
            resource: dictionary containing the resource or FHIR resource object to be uploaded to the server

        Returns:
            CreateResponse containing the server response and the resource that was uploaded

        """
        # parse the resource given as dictionary into a fhir resource model
        if isinstance(resource, dict):
            resource_type = resource.get("resourceType")
            if not resource_type:
                raise ValueError("No resource type defined in resource dictionary")
            resource = fhir.resources.construct_fhir_element(resource_type, resource)
        else:
            resource = resource.validate(resource)
        response = await self._upload_resource_async(resource)
        return ResourceCreateResponse(
            server_response_dict=dict(response.headers), resource=resource
        )

    def add_all(
        self,
        resources: List[Union[Resource, FHIRAbstractModel, dict]],
        batch_size: int = 1000,
        display: bool = True,
    ) -> BundleCreateResponse:
        """
        Upload a list of resources to the server, after packaging them into a bundle
        Args:
            resources: list of resources to upload to the server, either dictionary or FHIR resource objects
            batch_size: maximum number of resources to upload in one bundle
            display: whether to display a progress bar when the upload is batched

        Returns:
            Bundle create response from the fhir server

        """
        response = None
        if len(resources) > batch_size:
            # split the list of resources into batches of size batch_size
            batches = [
                resources[i : i + batch_size]
                for i in range(0, len(resources), batch_size)
            ]

            p_bar = tqdm(batches, disable=not display)
            for i, batch in enumerate(p_bar):
                p_bar.set_description(
                    f"Uploading Batch {i + 1}/{len(batches)}, n={len(batch)}"
                )
                bundle = make_transaction_bundle(
                    method=TransactionMethod.POST, resources=batch
                )
                if not response:
                    response = self._upload_bundle(bundle)
                else:
                    add_response = self._upload_bundle(bundle)
                    response.create_responses.extend(add_response.create_responses)
        else:
            bundle = make_transaction_bundle(
                method=TransactionMethod.POST, resources=resources
            )
            response = self._upload_bundle(bundle)
        return response

    async def add_all_async(
        self,
        resources: List[Union[Resource, FHIRAbstractModel, dict]],
        batch_size: int = 5000,
        display: bool = True,
    ) -> BundleCreateResponse:
        """
        Asynchronously upload a list of resources to the server, after packaging them into a bundle

        Args:
            resources: list of resources to upload to the server, either dictionary or FHIR resource objects
            batch_size: maximum number of resources to upload in one bundle
            display: whether to display a progress bar when the upload is batched

        Returns: Bundle create response from the fhir server
        """

        response = None
        if len(resources) > batch_size:
            # split the list of resources into batches of size batch_size
            batches = [
                resources[i : i + batch_size]
                for i in range(0, len(resources), batch_size)
            ]
            p_bar = tqdm(batches, disable=not display)
            for i, batch in enumerate(p_bar):
                p_bar.set_description(
                    f"Uploading Batch {i + 1}/{len(batches)}, n={len(batch)}"
                )

                # create a bundle from the batch of resources
                bundle = make_transaction_bundle(
                    method=TransactionMethod.POST, resources=batch
                )
                if not response:
                    response = await self._upload_bundle_async(bundle)
                else:
                    add_response = await self._upload_bundle_async(bundle)
                    response.create_responses.extend(add_response.create_responses)
        else:
            # create a bundle from the list of resources
            bundle = make_transaction_bundle(
                method=TransactionMethod.POST, resources=resources
            )
            response = await self._upload_bundle_async(bundle)
        return response

    def add_bundle(
        self, bundle: Union[Bundle, dict, str], validate: bool = True
    ) -> BundleCreateResponse:
        """
        Upload a bundle to the server
        :param bundle: str, dict or Bundle object to upload to the server
        :param validate: whether to validate the entries in the bundle
        :return: BundleCreateResponse from the fhir server containing the server assigned ids of the resources in
        the bundle
        """
        # create bundle and validate it
        if isinstance(bundle, dict):
            bundle = Bundle(**bundle)
        elif isinstance(bundle, str):
            bundle = Bundle.parse_raw(bundle)
        else:
            bundle = Bundle.validate(bundle)
        # check that all entries are bundle requests with methods post/put
        if validate:
            self._validate_upload_bundle_entries(bundle.entry)

        create_response = self._upload_bundle(bundle)
        return create_response

    async def add_bundle_async(
        self, bundle: Union[Bundle, dict, str], validate_entries: bool = True
    ) -> BundleCreateResponse:
        """
        Asynchronously upload a bundle to the server
        :param bundle: str, dict or Bundle object to upload to the server
        :param validate_entries: whether to validate the entries in the bundle
        :return: BundleCreateResponse from the fhir server containing the server assigned ids of the resources in
        the bundle
        """
        # create bundle and validate it
        if isinstance(bundle, dict):
            bundle = Bundle(**bundle)
        elif isinstance(bundle, str):
            bundle = Bundle.parse_raw(bundle)
        else:
            bundle = Bundle.validate(bundle)
        # check that all entries are bundle requests with methods post/put
        if validate_entries:
            self._validate_upload_bundle_entries(bundle.entry)

        transaction_response = await self._upload_bundle_async(bundle)
        return transaction_response

    def update(self, resources: List[Union[FHIRResourceModel, dict]]) -> dict:
        """
        Update a list of resources that exist on the server
        Args:
            resources: List of updated resources coming to send to the server

        Returns: Bundle update response from the fhir server

        """

        update_bundle = make_transaction_bundle(
            method=TransactionMethod.PUT, resources=resources
        )
        with self._sync_client() as client:
            r = client.post(self.api_address, json=json_dict(update_bundle))
            r.raise_for_status()
        return r.json()

    async def update_async(self, resources: List[Union[FHIRResourceModel, dict]]):
        update_bundle = make_transaction_bundle(
            method=TransactionMethod.PUT, resources=resources
        )

        async with self._async_client() as client:
            r = await client.post(self.api_address, json=json_dict(update_bundle))
            r.raise_for_status()
        return r.json()

    def delete(
        self,
        resources: List[Union[FHIRResourceModel, dict]] = None,
        references: List[Union[str, Reference]] = None,
        query: FhirQuerySync = None,
    ) -> None:
        """
        Delete resources from the server. Either resources, references or a query must be specified.
        Args:
            resources: Resources coming from the server containing an id to delete
            references: references {Resource}/{id} to delete
            query: query to use to find resources to delete

        Returns:
            None
        """

        self._validate_delete_args(query, references, resources)

        # if a query is specified, get the resources to delete
        if query:
            resp = query.all()
            resources = resp.resource_list
        delete_bundle = make_transaction_bundle(
            method=TransactionMethod.DELETE,
            resources=resources,
            references=references,
        )

        with self._sync_client() as client:
            r = client.post(self.api_address, json=json_dict(delete_bundle))
            r.raise_for_status()

    async def delete_async(
        self,
        resources: List[Union[FHIRResourceModel, dict]] = None,
        references: List[Union[str, Reference]] = None,
        query: FhirQueryAsync = None,
    ) -> None:
        """
        Asynchronously delete resources from the server. Either resources, references or a query must be specified.
        Args:
            resources: Resources coming from the server containing an id to delete
            references: references {Resource}/{id} to delete
            query: query to use to find resources to delete

        Returns:
            None

        """
        self._validate_delete_args(query, references, resources)
        # if a query is specified, get the resources to delete
        if query:
            resp = await query.all()
            resources = resp.resource_list

        delete_bundle = make_transaction_bundle(
            method=TransactionMethod.DELETE,
            resources=resources,
            references=references,
        )

        async with self._async_client() as client:
            r = await client.post(self.api_address, json=json_dict(delete_bundle))
            r.raise_for_status()

    def transfer(
        self,
        target_server: "FhirServer",
        query: FhirQuerySync = None,
        resources: List[Union[Resource, FHIRAbstractModel]] = None,
        get_missing: bool = True,
        record_linkage: bool = True,
        display: bool = False,
    ) -> TransferResponse:
        """
        Transfer resources from this server to another server while using server assigned ids and keeping referential
        integrity.

        Args:
            target_server: FhirServer to transfer to
            query: FhirQuerySync to use to find resources to transfer
            resources: list of resources to transfer
            get_missing: whether to get missing references from the source server
            record_linkage: whether to record the linkage between the source and target server
            display: whether to display the progress bar

        Returns:
            Transfer response for the transfer of the query result to the target server

        """

        response = transfer(
            source=self,
            target=target_server,
            query=query,
            resources=resources,
            get_missing=get_missing,
            record_linkage=record_linkage,
            display=display,
        )
        return response

    def summary(self, display: bool = True) -> ServerSummary:
        """
        Create a summary for the server. Contains resource counts for all resources available on the server.
        Args:
            display: whether to display a progress bar
        Returns:
            ServerSummary containing resource counts for all resources available on the server

        """
        summary = create_server_summary(self, self.rest_resources, display)
        return summary

    async def summary_async(self, display: bool = True) -> ServerSummary:
        """
        Asynchronously create a summary for the server. Contains resource counts for all resources available on the
        server.

        Args:
            display: whether to display a progress bar
        Returns:
            ServerSummary containing resource counts for all resources available on the server

        """
        summary = await create_server_summary_async(self, self.rest_resources, display)
        return summary

    @property
    def capabilities(self) -> CapabilityStatement:
        """
        Get the capabilities statement for the server
        """
        if not self._meta_data:
            self._get_meta_data()
        return CapabilityStatement(**self._meta_data)

    @property
    def rest_resources(self) -> List[str]:
        """
        Get the list of resources available on the server
        """
        return [capa.type for capa in self.capabilities.rest[0].resource]

    @property
    def headers(self):
        headers = {"Content-Type": "application/fhir+json"}
        if self._headers:
            headers.update(self._headers)
        return headers

    @property
    def auth(self):
        if self._auth:
            return self._auth
        elif self.username and self.password:
            return httpx.BasicAuth(username=self.username, password=self.password)
        elif self.token:
            return BearerAuth(self.token)

    @staticmethod
    def _validate_api_address(api_address: str) -> str:
        """
        Validate that api address is well formed and remove trailing / if present
        https://stackoverflow.com/a/7160778/3838313

        Args:
            api_address: base endpoint for the fhir rest api of the server

        Returns:
            the validated api address

        """

        regex = re.compile(
            r"^(?:http|ftp)s?://"  # http:// or https://
            r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"  # domain...
            r"localhost|"  # localhost...
            r"[A-Za-z0-9_-]*|"  # single word with hyphen/underscore for docker
            r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"  # ...or ip
            r"(?::\d+)?"  # optional port
            r"(?:/?|[/?]\S+)$",
            re.IGNORECASE,
        )
        if re.match(regex, api_address):
            if api_address[-1] == "/":
                api_address = api_address[:-1]
            return api_address

        else:
            raise ValueError(f"Malformed API URL: {api_address}")

    @staticmethod
    def _validate_delete_args(query, references, resources):
        if query and (resources or references):
            raise ValueError("Cannot specify both query and resources/references")
        if not (query or resources or references):
            raise ValueError("Must specify either query or resources/references")

    @staticmethod
    def _validate_upload_bundle_entries(entries: List[BundleEntry]):
        for i, entry in enumerate(entries):
            if entry.request.method.lower() not in ["post", "put"]:
                raise ValueError(f"Entry {i}:  method is not in [post, put]")

    def _upload_bundle(self, bundle: Bundle) -> BundleCreateResponse:
        """
        Upload a bundle to the server
        Args:
            bundle: transaction bundle to upload to the server

        Returns:
            BundleCreateResponse with the server assigned ids

        """
        with self._sync_client() as client:
            r = client.post(url=self.api_address, json=json_dict(bundle))
            try:
                r.raise_for_status()
            except Exception as e:
                print(r.text)
                raise e
        bundle_response = BundleCreateResponse(r, bundle)
        return bundle_response

    async def _upload_bundle_async(self, bundle: Bundle) -> BundleCreateResponse:
        """
        Asynchronously upload a bundle to the server
        Args:
            bundle: Bundle to upload to the server

        Returns:
            BundleCreateResponse with the server assigned ids
        """
        async with self._async_client() as client:
            r = await client.post(url=self.api_address, json=json_dict(bundle))
            try:
                r.raise_for_status()
            except Exception as e:
                print(r.text)
                raise e
        bundle_response = BundleCreateResponse(r, bundle)
        return bundle_response

    def _upload_resource(self, resource: Resource) -> httpx.Response:
        """
        Upload a resource to the server
        Args:
            resource: Fhir resource to upload

        Returns:
            httpx.Response from the server
        """
        url = self.api_address + "/" + resource.get_resource_type()
        with self._sync_client() as client:
            r = client.post(url=url, json=json_dict(resource))
            try:
                r.raise_for_status()
            except Exception as e:
                print(r.text)
                raise e
        return r

    async def _upload_resource_async(self, resource: Resource) -> httpx.Response:
        """
        Asynchronously upload a resource to the server
        Args:
            resource: Fhir resource to upload

        Returns:
            httpx.Response from the server
        """
        url = self.api_address + "/" + resource.get_resource_type()
        async with self._async_client() as client:
            r = await client.post(url=url, json=json_dict(resource))
            try:
                r.raise_for_status()
            except Exception as e:
                print(r.text)
                raise e
        return r

    def _get_meta_data(self):
        url = self.api_address + "/metadata"
        with self._sync_client() as client:
            r = client.get(url)
            try:
                r.raise_for_status()
            except Exception as e:
                print(r.text)
                raise e
        response = r.json()
        self._meta_data = response

    def _sync_client(self) -> httpx.Client:
        """Get a synchronous httpx client

        Returns:
            _httpx.Client: synchronous httpx client
        """

        transport = self._setup_transport()
        client = httpx.Client(
            headers=self.headers,
            auth=self.auth,
            proxies=self._proxies,
            timeout=self._timeout,
            transport=transport,
        )
        return client

    def _async_client(self) -> httpx.AsyncClient:
        transport = self._setup_transport(async_transport=True)
        client = httpx.AsyncClient(
            headers=self.headers,
            auth=self.auth,
            proxies=self._proxies,
            timeout=self._timeout,
            transport=transport,
        )
        return client

    def _setup_transport(
        self, async_transport: bool = False
    ) -> Union[RetryTransport, httpx.AsyncHTTPTransport, httpx.HTTPTransport]:
        """Setup the transport for the httpx client if retryable methods or status codes are set
        for the server the requests will be retried according to the configuration

        Args:
            async_transport: if True return an async transport
        """

        if self.retry_status_codes or self.retryable_methods:
            return RetryTransport(
                wrapped_transport=httpx.AsyncHTTPTransport()
                if async_transport
                else httpx.HTTPTransport(),
                max_attempts=self.max_attempts,
                backoff_factor=self.backoff_factor,
                retry_status_codes=self.retry_status_codes,
                retryable_methods=self.retryable_methods,
                jitter_ratio=self.jitter_ratio,
                max_backoff_wait=self.max_backoff_wait,
            )
        else:
            return (
                httpx.AsyncHTTPTransport() if async_transport else httpx.HTTPTransport()
            )

    def _get_oidc_token(self):
        # get a new token if it is expired or not yet set
        if not self.oauth_token or self.oauth_token.is_expired():
            client = OAuth2Client(
                client_id=self.client_id,
                client_secret=self.client_secret,
                token_endpoint_auth_method="client_secret_jwt",
            )
            client.register_client_auth_method(ClientSecretJWT(self.oidc_provider_url))
            token = client.fetch_token(self.oidc_provider_url)
            self.token = token["access_token"]
            self.oauth_token = token

    @staticmethod
    def _validate_query_input(
        resource: str, query_parameters: FhirQueryParameters, query_string: str
    ):
        if resource and query_parameters:
            raise ValueError("Cannot specify both a resource and query parameters")
        if query_string and query_parameters:
            raise ValueError("Cannot specify both a query string and query parameters")
        if resource and query_string:
            raise ValueError("Cannot specify both a resource and query string")
        if not resource and not query_string and not query_parameters:
            raise ValueError(
                "Must specify either a resource, query string or query parameters"
            )

    def __repr__(self):
        return f"FhirServer(api_address={self.api_address})"


def _api_address_from_env() -> str:
    # load FHIR_API_URL
    api_url = os.getenv("FHIR_API_URL")
    if not api_url:
        api_url = os.getenv("FHIR_SERVER_URL")
    if not api_url:
        raise EnvironmentError("No FHIR api address specified")
    return FhirServer._validate_api_address(api_url)



from enum import Enum
from typing import List, Union

from fhir.resources import FHIRAbstractModel, construct_fhir_element
from fhir.resources.bundle import Bundle, BundleEntry, BundleEntryRequest
from fhir.resources.reference import Reference
from fhir.resources.resource import Resource


class TransactionMethod(str, Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"


class TransactionType(str, Enum):
    TRANSACTION = "transaction"
    BATCH = "batch"


def make_transaction_bundle(
    transaction_type: TransactionType = TransactionType.TRANSACTION,
    method: Union[TransactionMethod, str] = TransactionMethod.POST,
    resources: Union[List[Resource], List[dict]] = None,
    references: Union[List[Reference], List[str]] = None,
) -> Bundle:
    """
    Create a transaction bundle based on the resources, references, transaction type, and method.
    Args:
        transaction_type:
        method:
        resources:
        references:

    Returns:

    """

    # validate input parameters
    _validate_transaction_input(method, references, resources)

    bundle = Bundle.construct()
    bundle.type = transaction_type.value
    if resources:
        if not (
            isinstance(resources[0], FHIRAbstractModel)
            or isinstance(resources[0], dict)
        ):
            raise ValueError(
                "Resources must be a list of FHIR resources or a list of dicts"
            )
        entries = [
            make_transaction_entry(method, resource=resource) for resource in resources
        ]

    else:
        entries = []
        for reference in references:
            if isinstance(reference, Reference):
                reference = reference.reference
            entries.append(make_transaction_entry(method, url=reference))

    bundle.entry = entries

    return Bundle(**bundle.dict(exclude_none=True))


def _validate_transaction_input(
    method: Union[TransactionMethod, str],
    references: Union[List[Reference], List[str]],
    resources: Union[List[Resource], List[dict]],
):
    if isinstance(method, str):
        method = TransactionMethod(method)
    if not resources and not references:
        raise ValueError("Either resources or references must be provided")
    if resources and references:
        raise ValueError("Only one of resources or references can be provided")
    if method in [TransactionMethod.POST, TransactionMethod.PUT]:
        # if method is POST or PUT, resources must be provided
        if not resources:
            raise ValueError("Resources must be provided for POST and PUT transactions")

    elif method == TransactionMethod.GET:
        if not references:
            raise ValueError("References must be provided for GET transactions")


def make_transaction_entry(
    method: Union[TransactionMethod, str],
    url: str = None,
    resource: Union[Resource, dict] = None,
) -> BundleEntry:
    """Create a transaction entry for a bundle based on the method, url, and resource.
    If only a resource is provided, the url will be constructed from the resource otherwise the given url will be used.

    Args:
        method: the method to use for the transaction one of GET, POST, PUT, DELETE
        url: optional relative url to use for the transaction
        resource: optional FHIR resource to use for the transaction

    Returns:
        The transaction entry
    """
    if isinstance(method, str):
        method = TransactionMethod(method)

    if not url and not resource:
        raise ValueError("Either url or resource must be provided.")

    entry = BundleEntry.construct()

    if resource:
        # construct the fhir element if the resource is a dict
        if isinstance(resource, dict):
            try:
                resource = construct_fhir_element(
                    resource.get("resourceType"), resource
                )
            except Exception as e:
                raise ValueError(
                    f"Unable to construct resource from dict: {resource} \n{e}"
                )
        # remove the id if the method is POST (create)
        if method == TransactionMethod.PUT and not resource.id:
            raise ValueError("PUT requires a resource with an id")
        if method == TransactionMethod.POST:
            if hasattr(resource, "id"):
                del resource.id
        _add_resource_to_entry(entry, resource)

    url = _get_transaction_url_for_method(method, url, resource)

    entry.request = BundleEntryRequest(**{"method": method.value, "url": url})

    return entry


def _get_transaction_url_for_method(
    method: TransactionMethod, url: str = None, resource: Resource = None
) -> str:
    """
    Get the url for a transaction entry based on the method, url, and resource.
    Args:
        method:
        url:
        resource:

    Returns:

    """
    if method == TransactionMethod.GET:
        if not url:
            url = resource.relative_path()
    elif method == TransactionMethod.PUT:
        if not resource:
            raise ValueError("PUT requires a resource")
        url = resource.relative_path()

    elif method == TransactionMethod.DELETE:
        if not url:
            url = resource.relative_path()
    else:
        if not url:
            url = resource.get_resource_type()

    return url


def _add_resource_to_entry(
    entry: BundleEntry, resource: Union[FHIRAbstractModel, dict]
):
    if isinstance(resource, FHIRAbstractModel):
        entry.resource = resource
    elif isinstance(resource, dict):
        entry.resource = construct_fhir_element(resource.get("resourceType"), resource)
    else:
        raise ValueError("Resource must be a FHIR resource or a dict")



import os

import pandas as pd
import pytest
from dotenv import find_dotenv, load_dotenv

from fhir_kindling import FhirServer
from fhir_kindling.serde.flatten import (
    flatten_resource,
    flatten_resources,
    flatten_response,
)


@pytest.fixture
def api_url():
    """
    Base api url and env vars
    """
    load_dotenv(find_dotenv())

    return os.getenv("FHIR_API_URL", "http://test.fhir.org/r4")


@pytest.fixture
def server(api_url):
    server = FhirServer(
        api_address=api_url,
        client_id=os.getenv("CLIENT_ID"),
        client_secret=os.getenv("CLIENT_SECRET"),
        oidc_provider_url=os.getenv("OIDC_PROVIDER_URL"),
    )
    return server


def test_flatten(server):
    query_resource = "Condition"
    search_param = "subject"
    query = server.query("Patient")
    query = query.include(query_resource, search_param, reverse=True)
    print(query.query_parameters)
    conditions = query.all()
    print(conditions)


def test_flatten_response(server):
    patients = server.query("Patient").limit(30)

    patients_df = flatten_response(patients)
    assert isinstance(patients_df, pd.DataFrame)
    assert len(patients_df) == 30


def test_flatten_resource(server):
    patient = server.query("Patient").limit(1).resources[0]
    flat_patient = flatten_resource(patient)
    print(flat_patient)

    condition = server.query("Condition").limit(1).resources[0]
    flat_condition = flatten_resource(condition)
    print(flat_condition)

    # observation = server.query("Observation").limit(1).resources[0]
    # flat_obs = flatten_resource(observation)


def test_resources_to_csv(server):
    patients = server.query("Patient").all().resources
    patients_df = flatten_resources(patients)
    patients_df.to_csv("patients.csv")

    if os.path.exists("patients.csv"):
        os.remove("patients.csv")

    conditions = server.query("Condition").all().resources
    conditions_df = flatten_resources(conditions)
    conditions_df.to_csv("conditions.csv")

    if os.path.exists("conditions.csv"):
        os.remove("conditions.csv")



from collections.abc import MutableMapping
from typing import List, Union

import pandas as pd
from fhir.resources import FHIRAbstractModel
from fhir.resources.fhirresourcemodel import FHIRResourceModel

from fhir_kindling.fhir_query.query_response import QueryResponse


def flatten(
    resources: Union[List[FHIRResourceModel], List[FHIRAbstractModel]] = None,
    response: QueryResponse = None,
    save: bool = False,
    path: str = None,
    display_progress: bool = False,
) -> Union[pd.DataFrame, List[pd.DataFrame], None]:
    """
    Flatten a list of resources or a query response into a dataframe/multiple dataframes.
    Args:
        resources: list of resources
        response: fhir kindling query response possibly containing multiple resources
        save: save the output to one or multiple csv files
        path: path to save the csv files
        display_progress: display a progress bar

    Returns: pandas dataframe with columns corresponding to the flattened resource keys

    """
    if resources and response:
        raise ValueError("Only one of resources or response can be provided")
    if resources:
        return flatten_resources(resources)
    elif response:
        return flatten_response(response)
    else:
        raise ValueError("Either resources or response must be provided")


def flatten_response(
    response: QueryResponse,
) -> Union[pd.DataFrame, List[pd.DataFrame]]:
    """
    Flatten a query response into a dataframe/multiple dataframes.
    Args:
        response: fhir kindling query response possibly containing multiple resources

    Returns: pandas dataframe with columns corresponding to the flattened resource keys

    """
    included_resources = response.included_resources
    # Flatten the included resources into a list of dataframes
    if included_resources:
        dfs = []
        for resource in included_resources:
            dfs.append(flatten_resources(resource.resources))

        dfs.insert(0, flatten_resources(response.resources))
        return dfs
    else:
        return flatten_resources(response.resources)


def flatten_resources(
    resources: Union[List[FHIRResourceModel], List[FHIRAbstractModel]]
) -> pd.DataFrame:
    """
    Flatten a list of resources of a single resource type into a dataframe.
    Args:
        resources: list of resources
    Returns: pandas dataframe with columns corresponding to the flattened resource keys

    """

    flat_resources = []
    for resource in resources:
        flat_resources.append(flatten_resource(resource))
    return pd.DataFrame.from_records(flat_resources)


def flatten_resource(resource: Union[FHIRResourceModel, FHIRAbstractModel]) -> dict:
    """
    Flatten a resource into a single level dictionary.
    Args:
        resource: Fhir resource to convert to single level dict

    Returns: dictionary with only single level keys containing the fields of the resource

    """
    resource_dict = resource.dict(exclude_none=True)
    flat_dict = flatten_dict(resource_dict)
    return flat_dict


def flatten_dict(
    d, parent_key: str = "", sep: str = "_", keys: List[str] = None
) -> dict:
    """
    Flatten a nested dictionary into a single level dictionary.

    Extension of https://stackoverflow.com/a/6027615/3838313 capable of handling lists
    Args:
        d: dictionary to flatten
        parent_key: parent key user for recursion
        sep: separator for the nested keys
        keys: list of keys to keep, all other keys will be skipped

    Returns:

    """
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        elif isinstance(v, list):
            for i, item in enumerate(v):
                if isinstance(item, MutableMapping):
                    items.extend(flatten_dict(item, f"{new_key}_{i}", sep=sep).items())
                else:
                    items.append((f"{new_key}_{i}", item))
        else:
            items.append((new_key, v))
    return dict(items)



import os
import time
from typing import Any, List, Tuple, Union

import pendulum
from tqdm.autonotebook import tqdm

from fhir_kindling import FhirServer
from fhir_kindling.benchmark.constants import BenchmarkOperations, DefaultQueries
from fhir_kindling.benchmark.data import generate_benchmark_data
from fhir_kindling.benchmark.figures import plot_benchmark_results
from fhir_kindling.benchmark.results import BenchmarkResults
from fhir_kindling.fhir_query.query_parameters import FhirQueryParameters
from fhir_kindling.fhir_server.transfer import (
    reference_graph,
    resolve_reference_graph,
)
from fhir_kindling.generators import (
    PatientGenerator,
)

N_ATTEMPTS = 20
BATCH_SIZE = 100


class ServerBenchmark:
    steps: List[BenchmarkOperations]
    queries: List[Tuple[str, Union[str, FhirQueryParameters]]]

    def __init__(
        self,
        servers: List[FhirServer],
        server_names: Union[List[str], None] = None,
        n_attempts: int = N_ATTEMPTS,
        batch_size: int = BATCH_SIZE,
        dataset_size: int = 1000,
        custom_queries: Union[
            List[Tuple[str, Union[str, FhirQueryParameters]]], None
        ] = None,
        steps: List[Union[str, BenchmarkOperations]] = None,
    ):
        """Initialize a benchmark object to test the performance of a set of servers.

        Args:
            servers: List of servers to benchmark.
            server_names: Optionally provide a list of names for the servers. Which will be used in the results.
                Defaults to the servers api address.
            n_attempts: The number of attempts for each operation. Defaults to N_ATTEMPTS.
            batch_size: The size of the batches for evaluating batch operations. Defaults to BATCH_SIZE.
            dataset_size: The number of base resources in the dataset. Defaults to 1000.
            custom_queries: A list of custom FHIR search queries to be included in the benchmark. Defaults to None.
            steps: Select a subset of the steps to run. Defaults to None.

        Raises:
            ValueError: If the number of server names does not match the number of servers.
            Or if steps or custom_queries are not valid.
        """
        self.servers = servers

        if server_names and len(server_names) != len(servers):
            raise ValueError(
                f"Invalid server names len={len(server_names)}, and len server = {len(servers)} "
                "When providing server_names, must provide one for each server"
            )
        self.server_names = server_names
        self._results = BenchmarkResults()
        self.benchmark_resources = {}
        self.n_attempts = n_attempts
        self.batch_size = batch_size
        self.dataset = None

        self._setup(queries=custom_queries, steps=steps)
        self.dataset_generator = generate_benchmark_data(n_patients=dataset_size)

    def _setup(
        self,
        queries: Union[List[Tuple[str, Union[str, FhirQueryParameters]]], None] = None,
        steps: List[Union[str, BenchmarkOperations]] = None,
    ):
        """Set up the steps and custom queries for the benchmark.

        Args:
            queries: User submitted list of queries to include. Defaults to None.
            steps: User submitted subset of benchmark steps to perform. Defaults to None.

        Raises:
            ValueError: If the operations/steps given are invalid
            ValueError: If any of the queries given is invalid
        """
        # if a list of steps is provided run only these steps, otherwise run all steps
        if steps:
            self.steps = []
            for s in steps:
                if isinstance(s, str):
                    self.steps.append(BenchmarkOperations(s))
                elif isinstance(s, BenchmarkOperations):
                    self.steps.append(s)
                else:
                    raise ValueError(
                        "Steps must be a list of strings or BenchmarkOperations objects"
                    )
        else:
            self.steps = [BenchmarkOperations(step) for step in BenchmarkOperations]

        # if custom queries are provided use them, otherwise use the default queries
        if queries:
            self.queries = []
            for q in queries:
                if isinstance(q, str):
                    self.queries.append(FhirQueryParameters.from_query_string(q))
                elif isinstance(q, FhirQueryParameters):
                    self.queries.append(q)
                else:
                    raise ValueError(
                        "Custom_queries must be a list of fhir query strings or FhirQueryParameters objects"
                    )
        else:
            self.queries = [
                FhirQueryParameters.from_query_string(qs)
                for qs in [dfq.value for dfq in DefaultQueries]
            ]

    def run_suite(
        self, progress: bool = True, save: bool = True, results_dir: str = None
    ):
        """Run the the test suite configured for this benchmark instance.
        By default the steps are: Dataset generation, single resource insert, batch insert,
        dataset upload, search, update and delete.

        Args:
            progress: Wether to visualize progress using a progress bar. Defaults to True.
            save: Save the results to file once the suite is finished. Defaults to True.
            results_dir: Directory in which to save the results. If None defaults to current working directory.
        """
        # generate benchmark data
        if BenchmarkOperations.GENERATE in self.steps:
            self.dataset = self.dataset_generator.generate(display=progress)
            # remove the generate step from the list of steps to run
            self.steps.remove(BenchmarkOperations.GENERATE)

        # run the benchmark for each server
        for i, server in tqdm(
            enumerate(self.servers),
            desc=f"Running bechmarks for {len(self.servers)} servers:",
            disable=not progress,
            leave=False,
        ):
            name = None
            if self.server_names:
                name = self.server_names[i]
            self._benchmark_server(server, progress=progress, name=name)

        self._results.set_completed(True)
        if save:
            self._save(path=results_dir)

    def _benchmark_server(self, server: FhirServer, progress: bool, name: str = None):
        """Run the benchmark suite for a single server

        Args:
            server: The server to run the benchmark against
            progress: Whether to display a progress bar
            name: Optional name for the server. Defaults to None.
        """
        # Iterate over the benchmark steps
        for step in tqdm(
            self.steps,
            desc=f"Server {server.api_address}",
            disable=not progress,
            leave=False,
        ):
            server_name = name if name else server.api_address
            if step == BenchmarkOperations.GENERATE:
                pass
            if step == BenchmarkOperations.INSERT:
                self._benchmark_insert(
                    server,
                    server_name=server_name,
                )
            elif step == BenchmarkOperations.DATASET_INSERT:
                self._upload_dataset(server, server_name=server_name)
            elif step == BenchmarkOperations.QUERY:
                self._benchmark_search(server, server_name=server_name)
            elif step == BenchmarkOperations.UPDATE:
                pass  # TODO
            elif step == BenchmarkOperations.DELETE:
                self._benchmark_delete(server, server_name=server_name)

    @property
    def results(self):
        """Retunrs the results of the benchmark if the benchmark has completed.

        Raises:
            Exception: If the benchmark has not completed

        Returns:
            The results of the benchmark
        """
        if not self._results.completed:
            raise Exception("Benchmark not completed")
        return self._results

    def plot(self):
        """Plot the results of the benchmark

        Returns:
            The plotly figure displaying the results
        """
        fig = plot_benchmark_results(self.results)
        return fig

    def _upload_dataset(self, server: FhirServer, server_name: str):
        """Upload the generated dataset to the server and track the time it takes.

        Args:
            server: The server to upload to
            server_name: Name for the server
        """
        # create temp copy of dataset
        dataset = self.dataset.copy(deep=True)

        ds_graph = reference_graph(dataset.resources)
        start_time = time.perf_counter()
        added_resources, linkage = resolve_reference_graph(
            ds_graph, server, True, False
        )
        total = time.perf_counter() - start_time
        resource_refs = [r.reference.reference for r in added_resources]
        self._add_resource_refs_for_tracking(server=server, refs=resource_refs)
        self._results.add_result(
            BenchmarkOperations.DATASET_INSERT,
            server_name,
            total,
        )

    def _save(self, path: str = None):
        """Save the benchmark results and figure to file

        Args:
            path: Where to save the resulst. Defaults to current working directory.
        """
        self.results.save(path)
        figure = self.plot()
        if not path:
            figure_path = os.getcwd()
            datestring = pendulum.now().to_iso8601_string().replace(":", "_")
            figure_path = os.path.join(figure_path, f"benchmark_{datestring}.png")
            figure.write_image(figure_path)

    def _benchmark_insert(self, server: FhirServer, server_name: str):
        self._benchmark_insert_single(server, server_name)
        self._benchmark_batch_insert(server, server_name)

    def _benchmark_update(self, server: FhirServer):
        self._benchmark_update_single(server)
        self._benchmark_batch_update(server)

    def _benchmark_insert_single(self, server: FhirServer, server_name: str):
        resources = PatientGenerator(n=self.n_attempts).generate()

        timings = []
        added_resources = []
        for resource in resources:
            start_time = time.perf_counter()
            result = server.add(resource)
            added_resources.append(result.reference)
            end_time = time.perf_counter()
            total_time = end_time - start_time
            timings.append(total_time)
        self._results.add_result(
            BenchmarkOperations.INSERT,
            server_name,
            timings,
        )
        # track added resources for each server
        added_resources = self.benchmark_resources.get(server.api_address, [])
        added_resources.extend(added_resources)
        self.benchmark_resources[server.api_address] = added_resources

    def _benchmark_batch_insert(self, server: FhirServer, server_name: str):
        timings = []
        # run multiple attempts for inserting resources
        for _ in range(self.n_attempts):
            resources = PatientGenerator(n=self.batch_size).generate()
            start_time = time.perf_counter()
            response = server.add_all(resources)
            elapsed_time = time.perf_counter() - start_time
            timings.append(elapsed_time)
            # track the added resources
            self._add_resource_refs_for_tracking(
                server, [r.reference.reference for r in response.create_responses]
            )

        self._results.add_result(
            BenchmarkOperations.BATCH_INSERT,
            server_name,
            timings,
        )

    def _add_resource_refs_for_tracking(
        self, server: FhirServer, refs: Union[List[Any], Any]
    ):
        server_resources = self.benchmark_resources.get(server.api_address, [])
        if isinstance(refs, list):
            server_resources.extend(refs)
        else:
            server_resources.append(refs)
        self.benchmark_resources[server.api_address] = server_resources

    def _benchmark_search(self, server: FhirServer, server_name: str):
        query_results = {}
        for query in self.queries:
            query_attempts = []
            for _ in range(self.n_attempts):
                start_time = time.perf_counter()
                server.query(query_parameters=query).all()
                elapsed_time = time.perf_counter() - start_time
                query_attempts.append(elapsed_time)
            query_results[query.to_query_string()] = query_attempts

        self._results.add_result(
            BenchmarkOperations.QUERY,
            server_name,
            results=query_results,
        )

    def _benchmark_delete(self, server: FhirServer, server_name: str):
        start_time = time.perf_counter()
        server_resource_refs = self.benchmark_resources.get(server.api_address, [])
        if not server_resource_refs:
            raise Exception("No resources to delete")
        # delete all resources
        server.delete(references=server_resource_refs)
        elapsed_time = time.perf_counter() - start_time

        self._results.add_result(
            BenchmarkOperations.DELETE,
            server_name,
            elapsed_time,
        )

    def _benchmark_update_single(self, server: FhirServer):
        pass

    def _benchmark_batch_update(self, server: FhirServer):
        pass



from .dataset import DatasetGenerator
from .field_generator import FieldGenerator
from .patient import PatientGenerator
from .resource_generator import FieldValue, GeneratorParameters, ResourceGenerator



import random
from typing import List, Tuple
from uuid import uuid4

import pandas as pd
import pendulum
from faker import Faker
from fhir.resources.humanname import HumanName
from fhir.resources.patient import Patient
from fhir.resources.reference import Reference

from fhir_kindling.generators.base import BaseGenerator
from fhir_kindling.serde.json import json_dict


class PatientGenerator(BaseGenerator):
    def __init__(
        self,
        n: int,
        age_range: Tuple[int, int] = None,
        gender_distribution: Tuple[float, float, float, float] = None,
        organisation: Reference = None,
        generate_ids: bool = False,
    ):
        self.resource = Patient
        self.n = n
        self.age_range = age_range
        self.gender_distribution = gender_distribution
        self._birthdate_range = None
        self.organisation = organisation
        self.generate_ids = generate_ids
        self.resources = None

    def generate(
        self,
        references: bool = False,
        generate_ids: bool = False,
        as_dict: bool = False,
    ):
        patients = self._generate()
        self.resources = patients

        if as_dict:
            resources = [json_dict(patient) for patient in patients]
            if self.n == 1:
                if references:
                    return resources[0], self._generate_references()[0]
                return resources[0]
            else:
                if references:
                    return resources, self._generate_references()
        else:
            if references and not self.generate_ids:
                raise ValueError("Cannot generate references without generating ids")
            elif references:
                return patients, self._generate_references()

        return patients

    def _generate(self):
        patients = []
        names = self._generate_patient_names(self.n)
        for name in names:
            patient = self._generate_patient_data(name=name)
            patients.append(patient)
        return patients

    def _generate_patient_data(self, name: Tuple[str, str]) -> Patient:
        first_name, last_name = name
        gender = random.choices(
            ["male", "female", "other", "unknown"],
            weights=self.gender_distribution
            if self.gender_distribution
            else [0.45, 0.45, 0.1, 0.0],
            k=1,
        )[0]

        name = HumanName(**{"family": last_name, "given": [first_name]})

        birthdate = self._generate_birthdate()
        patient_dict = {"gender": gender, "name": [name], "birthDate": birthdate}
        if self.organisation:
            patient_dict["managingOrganization"] = self.organisation

        patient_id = str(uuid4())
        patient_dict["id"] = patient_id

        return Patient(**patient_dict)

    @staticmethod
    def _generate_patient_names(n: int):
        fake = Faker()

        names = []
        for _ in range(n):
            split = fake.name().split(" ")
            family = split[-1]
            given = " ".join(split[:-1])
            names.append((given, family))
        return names

    def _generate_birthdate(self):
        if not self._birthdate_range:
            if self.age_range:
                if isinstance(self.age_range[0], int):
                    # generate age range from 18-101 years old
                    now = pendulum.now()
                    youngest = pd.to_datetime(
                        (
                            now - pendulum.duration(years=self.age_range[0])
                        ).to_date_string()
                    )
                    oldest = pd.to_datetime(
                        (
                            now - pendulum.duration(years=self.age_range[1])
                        ).to_date_string()
                    )
                else:
                    raise ValueError(
                        f"Unsupported type ({type(self.age_range[0])}) for generating patient ages."
                        f"Only integers are supported."
                    )
            else:
                # generate age range from 18-101 years old
                now = pendulum.now()
                youngest = pd.to_datetime(
                    (now - pendulum.duration(years=18)).to_date_string()
                )
                oldest = pd.to_datetime(
                    (now - pendulum.duration(years=101)).to_date_string()
                )

            self._birthdate_range = (
                pd.date_range(oldest, youngest, freq="D").strftime("%Y-%m-%d").tolist()
            )

        birthdate = random.choice(self._birthdate_range)
        return birthdate

    def _generate_references(self) -> List[Reference]:
        return [
            Reference(reference=f"Patient/{patient.id}") for patient in self.resources
        ]

    def __repr__(self):
        return (
            f"<PatientGenerator(n={self.n}, age_range={self.age_range}, "
            f"gender_distribution={self.gender_distribution}>"
        )



import json
from typing import List

from fhir.resources.bundle import Bundle
from fhir.resources.reference import Reference
from fhir.resources.resource import Resource
from httpx import Response


class CreateResponse:
    @staticmethod
    def _process_location_header(server_create_response: dict):
        location_header: str = server_create_response.get(
            "Location", server_create_response.get("location")
        )
        if not location_header:
            raise ValueError("Location field not found.")
        split_location = location_header.split("/")
        location = "/".join(split_location[:-2])
        version = int(split_location[-1])
        resource_id = split_location[-3]
        return resource_id, location, version


class ResourceCreateResponse(CreateResponse):
    location: str = None
    resource: Resource = None
    version: int = None
    resource_id: str = None
    reference: Reference = None

    def __init__(self, server_response_dict: dict, resource: Resource):
        self.resource = resource
        resource_id, location, version = self._process_location_header(
            server_response_dict
        )
        self.location = location
        self.resource_id = resource_id
        self.resource.id = resource_id
        self.reference = Reference(
            **{"reference": self.resource.get_resource_type() + "/" + resource_id}
        )

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}(resource_id={self.resource_id}, location={self.location},"
            f" version={self.version})>"
        )


class BundleCreateResponse:
    create_responses: List[ResourceCreateResponse] = None

    def __init__(self, server_response: Response, bundle: Bundle):
        self.create_responses = []
        for i, entry in enumerate(server_response.json()["entry"]):
            resource = bundle.entry[i].resource
            create_response = ResourceCreateResponse(entry["response"], resource)
            self.create_responses.append(create_response)

    @property
    def resources(self):
        return [r.resource for r in self.create_responses]

    @property
    def references(self):
        return [r.reference for r in self.create_responses]

    def __repr__(self):
        return (
            f"BundleCreateResponse(create_responses={self.create_responses[0]}...{self.create_responses[-1]}, "
            f"num_resources={len(self.create_responses)})"
        )


class TransferResponse:
    origin_server: str
    destination_server: str
    create_responses: List[ResourceCreateResponse]
    linkage: dict
    n_transferred: int

    def __init__(
        self,
        origin_server: str,
        destination_server: str,
        create_responses: List[ResourceCreateResponse],
        linkage: dict = None,
    ):
        self.origin_server = origin_server
        self.destination_server = destination_server
        self.create_responses = create_responses
        self.linkage = linkage
        self.n_transferred = len(create_responses)

    def save_linkage(self, filename: str):
        with open(filename, "w") as f:
            json.dump(self.linkage, f)

    def __str__(self):
        return (
            f"TransferResponse origin({self.origin_server}) -> destination ({self.destination_server})\n"
            f"n_transferred: {self.n_transferred}, linkage: {'stored' if self.linkage else 'not stored'}"
        )

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}(origin_server={self.origin_server},"
            f" destination_server={self.destination_server})"
        )


class UpdateResponse:
    # TODO: implement
    def __init__(self, server_response: Response):
        pass



# import os

# from fhir_kindling.benchmark.bench import ServerBenchmark
# from fhir_kindling.fhir_server.fhir_server import FhirServer


# def test_benchmark():
#     server_1 = FhirServer(api_address=os.getenv("FHIR_API_URL"))
#     server_2 = FhirServer(api_address=os.getenv("TRANSFER_SERVER_URL"))
#     benchmark = ServerBenchmark(servers=[server_1, server_2], n_attempts=2, dataset_size=10)
#     benchmark.run_suite(progress=False)

#     assert True



import json
import os
from typing import Dict, List, Union

import pendulum

from fhir_kindling.benchmark.constants import BenchmarkOperations


class BenchmarkResults:
    """Object to store, access and save the results of a benchmark."""

    def __init__(self):
        self.results = {}
        self._completed = False

    @property
    def completed(self):
        return self._completed

    def set_completed(self, completed: bool):
        self._completed = completed

    def save(self, path: str = None):
        """Save the results of the benchmark as a json file.

        Args:
            path: Directory to save the results. If None defaults to current working directory
            figure: Whether to save the figure as a png file. Defaults to True.

        Returns:
            None
        """

        if not self.completed:
            raise ValueError(
                "Benchmark results not completed yet and can not be saved."
            )

        if not path:
            path = os.getcwd()

        date_string = pendulum.now().to_iso8601_string().replace(":", "_")
        file_name = f"benchmark_results_{date_string}.json"
        file_path = os.path.join(path, file_name)
        with open(file_path, "w") as f:
            json.dump(self.results, f, indent=2)

    @property
    def insert_single(self) -> Union[List[float], Dict[str, List[float]]]:
        return self.results["insert"]

    @property
    def query(
        self,
    ) -> Dict[str, Dict[str, List[float]]]:
        """
        Return the query results as a dictionary of server names to a list of queries and their execution times.
        """
        return self.results[BenchmarkOperations.QUERY.value]

    @property
    def batch_insert(self) -> Union[List[float], Dict[str, List[float]]]:
        """Return the batch insert results as a dictionary of server names to a list of batch insert times."""
        return self.results[BenchmarkOperations.BATCH_INSERT.value]

    @property
    def dataset_insert(self) -> Union[List[float], Dict[str, List[float]]]:
        """Return the dataset insert results as a dictionary of server names to a list of dataset insert times."""
        return self.results[BenchmarkOperations.DATASET_INSERT.value]

    @property
    def update(self) -> Union[List[float], Dict[str, List[float]]]:
        """Return the update results as a dictionary of server names to a list of update times."""
        return self.results[BenchmarkOperations.UPDATE.value]

    @property
    def delete(self) -> Union[List[float], Dict[str, List[float]]]:
        """Return the delete results as a dictionary of server names to a list of delete times."""
        return self.results[BenchmarkOperations.DELETE.value]

    @property
    def batch_delete(self) -> Union[List[float], Dict[str, List[float]]]:
        """Return the batch delete results as a dictionary of server names to a list of batch delete times."""
        return self.results[BenchmarkOperations.BATCH_DELETE.value]

    @property
    def resources(self) -> Dict[str, Dict[str, List[float]]]:
        """Tracks the number of resources generated, queried, updated and deleted for each server."""
        pass

    def add_result(
        self,
        operation: Union[str, BenchmarkOperations],
        server: str,
        results: Union[List[float], dict],
    ):
        """Add results for a given operation and server.

        Args:
            operation: string or BenchmarkOperations enum value for the operation
            server: Name of the server
            results: the results to store
        """
        if isinstance(operation, str):
            operation = BenchmarkOperations(operation)

        operation_results = self.results.get(operation.value, {})
        operation_results[server] = results
        self.results[operation] = operation_results

    def __repr__(self) -> str:
        return f"BenchmarkResults(completed={self.completed})"



import pathlib
from enum import Enum
from typing import Dict, List, Optional, Union

import httpx
import orjson
from fhir.resources import FHIRAbstractModel
from fhir.resources.bundle import Bundle
from pydantic import BaseModel

from fhir_kindling.fhir_query.query_parameters import FhirQueryParameters


class OutputFormats(Enum):
    """
    Enum for the output formats.
    """

    JSON = "json"
    XML = "xml"


class DataframeFormat(Enum):
    LIST = "list"
    SINGLE = "single"


class IncludedResources(BaseModel):
    """
    Container for resources included in the query via the _include/_revinclude parameters.
    """

    resource_type: str
    resources: Optional[List[FHIRAbstractModel]] = None


class ResponseStatusCodes(str, Enum):
    OK = 200
    CREATED = 201
    NOT_FOUND = 404
    BAD_REQUEST = 400
    INTERNAL_SERVER_ERROR = 500


class QueryResponse:
    """
    Response object for the results of a FHIR query executed against a FHIR server.
    """

    def __init__(
        self,
        response: Union[httpx.Response, str, dict],
        query_params: FhirQueryParameters,
        output_format: OutputFormats = OutputFormats.JSON,
        limit: int = None,
        count: int = None,
    ):
        self.format = output_format
        self._limit = limit
        self.query_params = query_params
        self.resource = query_params.resource
        self._resources = None
        self._included_resources = {}
        self._bundle = None
        self.status_code: ResponseStatusCodes = None
        self.count = count

        # parse the response after the rest of the setup is complete
        self.response: Union[str, dict] = self._process_server_response(response)

    @property
    def resources(self) -> List[FHIRAbstractModel]:
        """
        List of primary resources returned by the server.

        Returns:
            List of FHIRResourceModel objects returned by the server.

        """
        if self.format == OutputFormats.XML:
            raise NotImplementedError("Resource parsing not supported for xml format")
        else:
            # return empty list if no resource matched the query
            if not self.response.get("entry"):
                return []
            if self._resources is None:
                self._extract_resources()
            return self._resources

    @property
    def included_resources(self) -> List[IncludedResources]:
        """
        Returns the list of resources included in the search.
        Returns:
            List of IncludeResources objects containing the resource type and the list of resources of this type
            included in the search

        """
        if self.format == OutputFormats.XML:
            raise NotImplementedError("Resource parsing not supported for xml format")
        else:
            # return empty list if no resource matched the query
            if not self.response.get("entry"):
                return []
            if not self.query_params.include_parameters:
                return []
            # parse the included resources if they don't exist
            if not self._included_resources:
                self._extract_resources()

            included = []
            for resource_type, resources in self._included_resources.items():
                included.append(
                    IncludedResources(resource_type=resource_type, resources=resources)
                )
            return included

    @property
    def resource_list(self) -> List[FHIRAbstractModel]:
        """
        Returns a list of all resources returned by the server including the primary resources and the included
        resources.
        Returns:
            List of FHIRResourceModel objects returned by the server.

        """
        resources = self.resources
        for included in self.included_resources:
            resources.extend(included.resources)
        return resources

    @property
    def total(self) -> int:
        """
        Returns the total number of resources returned by the query.
        Returns:
            Total number of resources matching the query.

        """
        return len(self.resource_list)

    def save(
        self, file_path: Union[str, pathlib.Path], output_format: str = "json"
    ) -> None:
        """
        Save the response to a file.
        Args:
            file_path: path to the file to save the response to.
            output_format: output format one of xml|json

        Returns:
            None
        """

        if isinstance(file_path, str):
            file_path = pathlib.Path(file_path)

        output_format = OutputFormats(output_format)
        # check that only xml queries can be saved as xml
        if self.format == OutputFormats.XML:
            if not output_format == OutputFormats.XML:
                raise NotImplementedError("XML query results can only be saved as XML")

        self.format = output_format
        if self.format == OutputFormats.XML:
            with open(file_path, "w") as f:
                f.write(self.response)

        elif self.format == OutputFormats.JSON:
            with open(file_path, "wb") as f:
                # dump the response as json using orjson and indent 2
                f.write(orjson.dumps(self.response, option=orjson.OPT_INDENT_2))

    def _extract_resources(self):
        """
        Parse the resources from the server response bundle. Split into included resources and resources that match the
        query exactly.
        Returns:

        """
        if not self._resources:
            self._resources = []
        for entry in Bundle(**self.response).entry:
            # add the directly queried resource to the resources list
            if entry.resource.resource_type == self.resource:
                self._resources.append(entry.resource)

            # process included resources
            elif entry.search.mode == "include":
                # get list of included resources based on type, if it does not exist yet return empty list
                included_resources = self._included_resources.get(
                    entry.resource.resource_type, []
                )
                # update the list with the entry and update the included resources dict
                included_resources.append(entry.resource)
                self._included_resources[
                    entry.resource.resource_type
                ] = included_resources

    def _process_server_response(
        self, response: Union[httpx.Response, str, dict]
    ) -> Union[Dict, Bundle, str]:
        """
        Handle the initial response from the server and resolve pagination if necessary.
        Args:
            response: the initial server response

        Returns:
            the server response, with pagination resolved if necessary

        """

        # if the format is xml resolve pagination and return xml string response
        if self.format == OutputFormats.XML:
            if isinstance(response, httpx.Response):
                response = response.content
            return response
        # otherwise, resolve json pagination and process further according to selected outcome
        else:
            if isinstance(response, httpx.Response):
                response = response.json()
            elif isinstance(response, str):
                response = Bundle.parse_raw(response).dict()

            return response

    def __repr__(self):
        if self.format == OutputFormats.XML:
            return f"<QueryResponse(resource={self.resource}, format=xml)>"
        if self._included_resources:
            resources = list(self._included_resources.keys())
            return (
                f"<QueryResponse(resource={self.resource}, format=json, "
                f"included_resources={resources})>"
            )
        return f"<QueryResponse(resource={self.resource}, n={len(self.resources)})>"



from typing import List, Tuple, Type, Union

from fhir.resources import FHIRAbstractModel, get_fhir_model_class
from fhir.resources.fhirtypes import ResourceType
from fhir.resources.resource import Resource
from pydantic.fields import ModelField


def get_resource_fields(
    resource: Union[Resource, ResourceType, str, Type[FHIRAbstractModel]]
) -> List[ModelField]:
    """
    Get the fields of a resource.
    Args:
        resource:

    Returns:

    """
    # If the resource is a string, then we need to get the resource class
    if isinstance(resource, str):
        resource = get_fhir_model_class(resource)

    # Get the resource model fields
    fields = list(resource.__fields__.values())
    return fields


def valid_resource_name(
    resource_name: str, strict: bool = True
) -> Union[str, Tuple[str, bool]]:
    """
    Checks if the given resource name is a valid FHIR resource name.
    Args:
        resource_name:

    Returns:

    """
    try:
        get_fhir_model_class(resource_name)
        if strict:
            return resource_name
        return resource_name, True
    except KeyError:
        if strict:
            raise ValueError(f"Invalid resource name: {resource_name}")
        return resource_name, False


def check_resource_contains_field(
    resource: Union[Resource, ResourceType, str, Type[FHIRAbstractModel]],
    field_name: str,
):
    """
    Checks if the given resource contains the given field.
    Args:
        resource: Name of the resource to check.
        field_name: field name to check for.

    Returns:

    Raises:
        ValueError: If the resource does not contain the field.

    """

    fields = get_resource_fields(resource)
    field_names = [field.name for field in fields]
    if field_name not in field_names:
        raise ValueError(f"Resource {resource} does not contain field {field_name}")



import pandas as pd

from fhir_kindling.privacy.k_anonymity import is_k_anonymized


def test_k_anonymity(fhir_server):
    test_data = []
    for i in range(100):
        if i % 4 == 0:
            test_data.append({"name": "John", "age": i % 4})
        elif i % 4 == 1:
            test_data.append({"name": "Jack", "age": i % 4})
        elif i % 4 == 2:
            test_data.append({"name": "Jones", "age": i % 4})
        elif i % 4 == 3:
            test_data.append({"name": "Jill", "age": i % 4})

    df = pd.DataFrame(test_data)

    assert is_k_anonymized(df, k=3)

    non_anon_df = df.copy()
    non_anon_df["age"] = list(range(100))
    assert not is_k_anonymized(non_anon_df, k=3)

    assert is_k_anonymized(non_anon_df, k=3, excluded_cols=["age"])

    assert is_k_anonymized(non_anon_df, k=3, id_cols=["name"])


# def test_anonymize(fhir_server):
#     patients = fhir_server.query("Patient").limit(1000).resources
#
#     print(len(patients))
#     df = flatten_resources(patients)
#     anonymize(df)



from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

# from fhir_kindling.fhir_server.fhir_server import FhirServer
import tqdm
from pydantic import BaseModel

from fhir_kindling.util.resources import valid_resource_name

if TYPE_CHECKING:
    from fhir_kindling.fhir_server import FhirServer


class ResourceSummary(BaseModel):
    resource: str
    count: int


class ServerSummary(BaseModel):
    name: Optional[str] = None
    resources: List[ResourceSummary]

    @property
    def available_resources(self) -> List[ResourceSummary]:
        return [r for r in self.resources if r.count > 0]


def create_server_summary(
    server: "FhirServer", resources: List[str], display: bool = True
) -> ServerSummary:
    """
    Create a summary of the server's resources and counts.
    Args:
        server: FhirServer object to summarize
        resources: list of resources to query/count
        display: whether to display a progress bar

    Returns:
        ServerSummary object containing a list of ResourceSummary objects
    """
    resource_summaries = []
    summary = {"name": server.api_address}

    pbar = tqdm.tqdm(resources, disable=not display)
    for resource in pbar:
        pbar.set_description(f"Querying {resource}")
        resource, valid = valid_resource_name(resource, strict=False)
        if not valid:
            continue
        count = server.query(resource).count()
        resource_summaries.append(ResourceSummary(resource=resource, count=count))
    summary["resources"] = resource_summaries
    return ServerSummary(**summary)


async def create_server_summary_async(
    server: "FhirServer", resources: List[str], display: bool = True
) -> ServerSummary:
    """
    Asynchronously create a summary of the server's resources and counts.

    Args:
        server: FhirServer object to summarize
        resources: list of resources to query/count
        display: whether to display a progress bar

    Returns:
        ServerSummary object containing a list of ResourceSummary objects
    """
    resource_summaries = []
    summary = {"name": server.api_address}

    pbar = tqdm.tqdm(resources, disable=not display)
    for resource in pbar:
        pbar.set_description(f"Querying {resource}")
        resource, valid = valid_resource_name(resource, strict=False)
        if not valid:
            continue
        count = await server.query_async(resource).count()
        resource_summaries.append(ResourceSummary(resource=resource, count=count))
    summary["resources"] = resource_summaries
    return ServerSummary(**summary)



from __future__ import annotations

from typing import TYPE_CHECKING, List, OrderedDict, Tuple, Union

import networkx as nx
import orjson
from fhir.resources import FHIRAbstractModel, construct_fhir_element
from fhir.resources.resource import Resource
from pydantic import ValidationError
from tqdm.autonotebook import tqdm

from fhir_kindling.fhir_query import FhirQuerySync
from fhir_kindling.fhir_server.server_responses import (
    ResourceCreateResponse,
    TransferResponse,
)
from fhir_kindling.util.references import check_missing_references, extract_references

if TYPE_CHECKING:
    from fhir_kindling.fhir_server import FhirServer


def transfer(
    source: "FhirServer",
    target: "FhirServer",
    resources: List[Union[Resource, FHIRAbstractModel]] = None,
    query: FhirQuerySync = None,
    get_missing: bool = True,
    record_linkage: bool = True,
    display: bool = True,
) -> TransferResponse:
    """
    Transfer a list of resources from one server to another.

    Args:
        source: The server to transfer the resources from.
        target: The server to transfer the resources to.
        resources: A list of resources to transfer.
        query: A FhirQuerySync object to use to query the source server.
        get_missing: Whether to get missing resources from the source server.
        record_linkage: Whether to record the linkage between the source and target resources.
        display: Whether to display a progress bar.
    """

    # get the resources to transfer, including missing references
    transfer_resources = _get_transfer_resources(source, resources, query, get_missing)
    transfer_graph = reference_graph(transfer_resources)

    # process the graph to create the resources on the target server
    create_responses, linkage = resolve_reference_graph(
        transfer_graph, target, record_linkage, display
    )

    return TransferResponse(
        origin_server=source.api_address,
        destination_server=target.api_address,
        create_responses=create_responses,
        linkage=linkage,
    )


def reference_graph(resources: List[Union[Resource, FHIRAbstractModel]]) -> nx.DiGraph:
    """
    Creates a graph of the references in a list of resources.

    Args:
        resources: List of resource to create the graph from.

    Returns:
        A directed graph depicting the references in the resources.
    """
    dg = nx.DiGraph()
    for resource in resources:
        path = resource.relative_path()
        if path in dg:
            dg.nodes[path]["resource"] = resource
        else:
            dg.add_node(path, resource=resource)
        for reference in extract_references(resource):
            reference_path = f"{reference[1]}/{reference[2]}"
            if reference_path not in dg:
                dg.add_node(reference_path, resource=None)
            dg.add_edge(
                reference_path, path, field=reference[0], list_field=reference[3]
            )
    return dg


def resolve_reference_graph(
    graph: nx.DiGraph,
    target: "FhirServer",
    record_linkage: bool,
    display: bool,
) -> Tuple[List[ResourceCreateResponse], dict]:
    nodes = list(graph.nodes)

    linkage = {}
    create_responses = []
    # iterate over the graph and update the references off all successors node to match the newly created resources
    # on the target server

    with tqdm(total=len(nodes), disable=not display) as pbar:
        while len(nodes) > 0:
            top_nodes = [
                node for node in nodes if len(list(graph.predecessors(node))) == 0
            ]

            # get the resources from the graph

            resources = []

            for node in top_nodes:
                try:
                    resources.append(_resource_from_graph_node(graph, node))
                except Exception as e:
                    print(e)
                    print(node)
                    raise e
            # insert the resources into the target server
            create_response = target.add_all(resources=resources)

            # iterate through the top nodes and update the successors with the references from the target server
            for node, reference in zip(top_nodes, create_response.references):
                if record_linkage:
                    hash_origin = hash(node)
                    linkage[hash_origin] = reference.reference
                _update_successors(graph, node, reference.reference)

            # add the create response to the list of responses
            create_responses.extend(create_response.create_responses)

            # remove the processed top nodes from the graph
            graph.remove_nodes_from(top_nodes)
            nodes = list(graph.nodes)
            pbar.update(len(top_nodes))

    return create_responses, linkage


def _update_successors(graph: nx.DiGraph, node: str, reference: str):
    """
    Update the successors of a node in a graph with the updated reference from the new server.

    Args:
        graph: The graph to update.
        node: The node to update.
        reference: The reference to update the node with.
    """
    for successor in graph.successors(node):
        field = graph[node][successor]["field"]
        list_field = graph[node][successor]["list_field"]
        resource = graph.nodes[successor]["resource"]

        if list_field:
            # Find the item that references the node
            reference_list = graph.nodes[successor]["resource"].dict()[field]
            reference_item = next(
                (item for item in reference_list if item.get("reference") == str(node)),
                None,
            )
            if reference_item:
                # update the resource with the new reference
                index = reference_list.index(reference_item)
                resource = orjson.loads(resource.json())
                resource[field][index] = {"reference": reference}
                graph.nodes[successor]["resource"] = resource
            else:
                print("Reference item not found")
        else:
            if not isinstance(graph.nodes[successor]["resource"], dict):
                resource = orjson.loads(resource.json())
                resource[field] = {"reference": reference}
                graph.nodes[successor]["resource"] = resource
            else:
                graph.nodes[successor]["resource"][field] = reference


def _resource_from_graph_node(graph: nx.DiGraph, node: str) -> FHIRAbstractModel:
    """
    Get a resource from a graph node.

    Args:
        graph: The graph to get the resource from.
        node: The node to get the resource from.

    Returns:
        The resource at the node.
    """
    resource = graph.nodes[node]["resource"]

    if resource:
        if isinstance(resource, OrderedDict) or isinstance(resource, dict):
            resource_dict = dict(resource)
            resource_type = resource_dict.get(
                "resourceType", resource_dict.get("resource_type")
            )
            resource_json = orjson.dumps(resource_dict)
            try:
                resource = construct_fhir_element(resource_type, resource_json)
            except ValidationError as e:
                print(f"Error creating resource: {e}")
                print(f"Resource: {resource_dict}")
                raise e
            except Exception as e:
                print(f"Error creating resource: {e}")
                print(f"Resource: {resource_dict}")
                raise e
        elif isinstance(resource, FHIRAbstractModel):
            pass
        else:
            raise ValueError(f"Unknown resource type: {type(resource)}")
        return resource
    else:
        raise ValueError(f"Resource not found for node {node}")


def _get_transfer_resources(
    source: "FhirServer",
    resources: List[Union[Resource, FHIRAbstractModel]] = None,
    query: FhirQuerySync = None,
    get_missing: bool = True,
) -> List[FHIRAbstractModel]:
    if query and resources:
        raise ValueError("Cannot specify both query and resources")
    if not query and not resources:
        raise ValueError(
            f"Must specify either query or resources. Query: {query}, Resources: {resources}"
        )
    # if query parameters are given execute the query against the server
    if query:
        resources = query.all().resource_list
    # find missing references
    missing_references = check_missing_references(resources)

    # get missing references
    if missing_references:
        if get_missing:
            missing = source.get_many(missing_references)
            resources.extend(missing)
        else:
            raise ValueError(
                f"Related resources of the resources to be transferred are missing:\n{missing_references} \n\n"
                f"To get these resources, set get_missing=True."
            )

    return resources



import random
from typing import Dict, List, Optional, Type, Union
from uuid import uuid4

import matplotlib.pyplot as plt
import networkx as nx
from fhir.resources import (
    FHIRAbstractModel,
    construct_fhir_element,
    get_fhir_model_class,
)
from fhir.resources.fhirresourcemodel import FHIRResourceModel
from fhir.resources.fhirtypes import ReferenceType
from fhir.resources.reference import Reference
from pydantic import BaseModel
from tqdm.autonotebook import tqdm

from fhir_kindling.fhir_server import FhirServer
from fhir_kindling.fhir_server.transfer import reference_graph, resolve_reference_graph
from fhir_kindling.generators.base import BaseGenerator
from fhir_kindling.generators.patient import PatientGenerator
from fhir_kindling.generators.resource_generator import ResourceGenerator
from fhir_kindling.generators.time_series_generator import TimeSeriesGenerator
from fhir_kindling.util import get_resource_fields


class DataSetResourceGenerator(BaseGenerator):
    name: str
    generator: BaseGenerator
    depends_on: Optional[Union[str, List[str]]] = None
    reference_field: Optional[
        Union[str, List[str], List[Union[str, None]], None]
    ] = None
    likelihood: float
    references: Dict[str, Reference]

    def __init__(
        self,
        name: str,
        generator: BaseGenerator,
        depends_on: Optional[Union[str, List[str]]] = None,
        reference_field: Optional[
            Union[str, List[str], List[Union[str, None]], None]
        ] = None,
        likelihood: float = 1.0,
    ):
        self.name = name
        self.generator = generator
        self.depends_on = depends_on
        self.reference_field = reference_field
        self.likelihood = likelihood
        self.references = {}

    def add_reference(self, reference_field: str, reference: Union[Reference, str]):
        if isinstance(reference, str):
            reference = {"reference": reference}
        elif isinstance(reference, Reference):
            reference = reference.dict()
        else:
            raise ValueError(
                f"Reference must be a string or a Reference object, got {type(reference)}"
            )

        self.references[reference_field] = reference

    def generate(self):
        # generate based on the likelihood
        if self.likelihood < 1.0:
            if random.random() > self.likelihood:
                return None

        if isinstance(self.generator, TimeSeriesGenerator):
            return self._generate_time_series()

        elif isinstance(self.generator, ResourceGenerator) or isinstance(
            self.generator, PatientGenerator
        ):
            return self._generate_single()

        else:
            raise ValueError(
                f"Expected ResourceGenerator or TimeSeriesGenerator, got {type(self.generator)}"
            )

    def _generate_single(self):
        base_resource_dict = self.generator.generate(generate_ids=True, as_dict=True)
        if self.references:
            # insert the references

            base_resource_dict = {**base_resource_dict, **self.references}

        resource = self.generator.resource(**base_resource_dict)
        return resource

    def _generate_time_series(self):
        resources = self.generator.generate(generate_ids=True, as_dict=True)
        if self.references:
            # insert the references
            for r in resources:
                for field, reference in self.references.items():
                    if isinstance(reference, str):
                        reference = {"reference": reference}
                    r[field] = reference
            # resources = [{**resource, **self.references} for resource in resources]
        if isinstance(self.generator, TimeSeriesGenerator):
            r_type = self.generator.generator.resource
            return [r_type(**resource) for resource in resources]
        else:
            raise ValueError(
                f"TimeSeriesGenerator expected, got {type(self.generator)}"
            )

    def __repr__(self) -> str:
        return f"<DataSetResourceGenerator {self.name}, generator={self.generator}>"


class DataSet(BaseModel):
    name: str
    base_resource: str
    resources: List
    resource_types: List[str]

    @property
    def n_resources(self):
        return len(self.resources)

    def size(self, human_readable: bool = False):
        """The size of the dataset in bytes"""

        size = sum(
            [
                len(resource.json(exclude_none=True, return_bytes=True))
                for resource in self.resources
            ]
        )
        if human_readable:
            return size / 1024 / 1024
        return size

    def upload(self, server: "FhirServer", display: bool = False):
        ds_graph = reference_graph(self.resources)
        result, _ = resolve_reference_graph(ds_graph, server, True, display=display)
        return result


class DatasetGenerator:

    """
    Generates a dataset of FHIR resources.
    """

    def __init__(self, base_resource: str = "Patient", n: int = None, name: str = None):
        self.name = name if name else str(uuid4())

        if base_resource == "Patient":
            self.base_resource = get_fhir_model_class(base_resource)
        else:
            raise NotImplementedError(
                "DatasetGenerator not implemented with {} as a base resource".format(
                    base_resource
                )
            )
        self.n = n
        self.generators: List[DataSetResourceGenerator] = []
        self._references = None
        self._patients = None
        self._dataset: DataSet = None
        self._nodes = set()
        self._graph = nx.DiGraph()
        self._resource_types = set()

        self.setup()

    def setup(self):
        # add base generator
        self.add_resource_generator(
            PatientGenerator(generate_ids=True, n=1),
            name="base",
            depends_on=None,
            reference_field=None,
            likelihood=1.0,
        )

    def generate(self, display: bool = False) -> DataSet:
        """
        Generate a dataset of FHIR resources according to the given conditions

        Args:
            ids:

        Returns:

        """
        resources = []
        for _ in tqdm(range(self.n), disable=not display, desc="Generating dataset"):
            batch = self._generate_resources_from_graph()
            added_resources = []

            for k, v in batch.items():
                if v is None:
                    continue

                if isinstance(v, list):
                    added_resources.extend(v)
                else:
                    added_resources.append(v)
            resources.extend(added_resources)

        dataset = self._make_data_set(resources)
        self._dataset = dataset
        return dataset

    def _make_data_set(self, resources: List[dict]) -> DataSet:
        # construct fhir elements
        fhir_resources = []
        for resource in resources:
            if isinstance(resource, dict):
                resource = construct_fhir_element(resource["resourceType"], resource)
            else:
                fhir_resources.append(resource)
        dataset = DataSet(
            name=self.name,
            base_resource=self.base_resource.get_resource_type(),
            resources=fhir_resources,
            resource_types=list(self._resource_types),
        )

        return dataset

    def _generate_resources_from_graph(self):
        """
        Generate a set of resource based on the generator graph
        """

        results = {}

        graph = self.graph()
        # start with the base node/generator and work through the graph
        top_list = reversed(list(nx.topological_sort(graph)))
        for node in top_list:
            # get the edges from the node
            if not self._check_dependencies(node, results):
                continue
            # get the generator from the node data
            generator = self._get_node_generator(node)

            # get the dependencies and add them to the generator
            self._get_refs_for_generator(generator, results)

            result = generator.generate()
            results[node] = result
            # print("Generated", node, result[node])
        return results

    def _get_refs_for_generator(
        self, generator: DataSetResourceGenerator, result: dict
    ) -> dict:
        """Get the references resources for a generator from the results and insert a reference to them into
        the reference field of the generator

        Args:
            generator: the generator to get the references for
            result: dependencies that have already been generated

        Returns:
            None
        """
        if not generator.reference_field:
            return

        if isinstance(generator.reference_field, list):
            for dep, ref_field in zip(generator.depends_on, generator.reference_field):
                if not ref_field:
                    continue
                resource = result.get(dep)
                if resource is None:
                    raise ValueError(f"Resource {dep} is None")

                ref = self._get_ref_from_resource_dict(resource)
                generator.add_reference(reference=ref, reference_field=ref_field)
        else:
            resource = result.get(generator.depends_on)
            if resource is None:
                raise ValueError(f"Resource dependency {generator.depends_on} is None")
            ref = self._get_ref_from_resource_dict(resource)
            generator.add_reference(
                reference=ref, reference_field=generator.reference_field
            )

    def _get_ref_from_resource_dict(self, resource_dict: dict) -> str:
        """Get a reference dict from a resource dict by combining the resource type and id

        Args:
            resource_dict: dictionary representation of a fhir resource

        Returns:
            Dictionary representation of a fhir reference
        """

        if not isinstance(resource_dict, dict):
            resource_dict = resource_dict.dict()
        ref_string = "{}/{}".format(resource_dict["resourceType"], resource_dict["id"])

        return ref_string

    def _get_node_generator(self, node: str) -> DataSetResourceGenerator:
        """
        Get a generator by name

        Args:
            node: the name of the generator

        Returns:
            the generator
        """
        try:
            generator = self.graph().nodes[node]["generator"]
            return generator
        except KeyError:
            raise ValueError("No generator found for node {}".format(node))

    def _check_dependencies(self, node: str, result: dict) -> bool:
        """
        Check if a node has any dependencies

        Args:
            node: the node to check
        """

        graph = self.graph()
        edges = graph.edges(node, data=True)
        if not edges:
            return True

        dependencies_exists = True
        for edge in edges:
            _, dependency, data = edge
            if not result.get(dependency):
                dependencies_exists = False
                break

        return dependencies_exists

    def graph(self) -> nx.DiGraph:
        """
        Return a networkx graph of the descripting the resource generation process
        """
        if not self._graph:
            self._graph = self._generate_resource_graph()
        return self._graph

    def explain(self):
        figure = self.draw_graph()
        figure.show()

    def draw_graph(self):
        """
        Visualize the graph underlying the resource generation process
        """
        graph = self.graph()
        pos = nx.spring_layout(graph)
        edge_labels = nx.get_edge_attributes(graph, "likelihood")
        nx.draw_networkx_edge_labels(graph, pos, edge_labels=edge_labels)
        nx.draw(graph, pos, with_labels=True)
        return plt

    def add_resource_generator(
        self,
        resource_generator: BaseGenerator,
        name: str,
        depends_on: Union[str, List[str]] = "base",
        reference_field: Union[str, List[str], None] = None,
        likelihood: float = 1.0,
    ) -> "DatasetGenerator":
        """
        Adds a resource generator to the dataset generator.

        Args:
            resource_generator (ResourceGenerator): The resource generator to add.
            name (str): The name of the generator.
            depends_on (Union[str, List[str]], optional): The name(s) of the generator(s) that this generator
            depends on. Defaults to "base".
            reference_field (Union[str, List[str], None], optional): The name(s) of the reference field(s)
            that this generator uses to reference other resources. Defaults to None.
            likelihood (float, optional): The likelihood of generating this resource. Defaults to 1.0.

        Returns:
            DatasetGenerator: The dataset generator instance.

        Raises:
            ValueError: If a generator with the same name already exists.
        """

        # make sure that the node names are unique
        if name in self._nodes:
            raise ValueError("A generator with the name {} already exists".format(name))
        else:
            self._nodes.add(name)

        # validate the reference field
        self._validate_depends_and_reference(depends_on, reference_field)

        if isinstance(resource_generator, ResourceGenerator):
            self._resource_types.add(resource_generator.resource.get_resource_type())
        elif isinstance(resource_generator, TimeSeriesGenerator):
            self._resource_types.add(
                resource_generator.generator.resource.get_resource_type()
            )
        elif isinstance(resource_generator, PatientGenerator):
            self._resource_types.add("Patient")
        else:
            raise ValueError(
                "Resource generator must be of type ResourceGenerator or TimeSeriesGenerator"
                "got {}".format(type(resource_generator))
            )

        generator = DataSetResourceGenerator(
            name=name,
            generator=resource_generator,
            depends_on=depends_on,
            reference_field=reference_field,
            likelihood=likelihood,
        )

        # add the generator to the graph
        self.generators.append(generator)
        self._add_generator_to_graph(generator)

        return self

    def _validate_depends_and_reference(
        self,
        depends_on: Union[str, List[str]],
        reference_field: Union[str, List[str], None],
    ):
        self._validate_depends(depends_on)

        if reference_field:
            if isinstance(depends_on, list) and isinstance(reference_field, list):
                if len(depends_on) != len(reference_field):
                    raise ValueError(
                        "When provided as list the number of reference fields must match the number of dependencies"
                    )
            elif isinstance(depends_on, list) and isinstance(reference_field, str):
                raise ValueError(
                    "When provided as list the number of reference fields must match the number of dependencies"
                )
            elif isinstance(depends_on, str) and isinstance(reference_field, str):
                pass
            else:
                raise ValueError(
                    "When provided as list the number of reference fields must match the number of dependencies"
                )

    def _validate_depends(
        self,
        depends_on: Union[str, List[str]],
    ):
        graph = self.graph()
        if depends_on:
            if isinstance(depends_on, list):
                for dependency in depends_on:
                    if dependency not in graph.nodes:
                        raise ValueError(
                            "The dependency {} does not exist in the graph".format(
                                dependency
                            )
                        )
            else:
                if depends_on not in graph.nodes:
                    raise ValueError(
                        "The dependency {} does not exist in the graph".format(
                            depends_on
                        )
                    )

    def _add_generator_to_graph(self, generator: DataSetResourceGenerator):
        self._graph.add_node(generator.name, generator=generator)
        if generator.depends_on:
            if isinstance(generator.depends_on, str):
                self._graph.add_edge(
                    generator.name,
                    generator.depends_on,
                    likelihood=generator.likelihood,
                    reference_field=generator.reference_field,
                )
            elif isinstance(generator.depends_on, list):
                for i, depends in enumerate(generator.depends_on):
                    self._graph.add_edge(
                        generator.name,
                        depends,
                        likelihood=generator.likelihood,
                        reference_field=generator.reference_field[i]
                        if generator.reference_field
                        else None,
                    )
            else:
                raise ValueError(
                    "depends_on must be either a string referencing a node or a list of strings "
                    "of node references. Got {}".format(type(generator.depends_on))
                )

    def _generate_resource_graph(self) -> nx.DiGraph:
        graph = nx.DiGraph()
        for generator in self.generators:
            self._add_generator_to_graph(generator)
        return graph

    def _add_reference_param(self, resource: FHIRResourceModel, reference: Reference):
        # check if a reference field is present for the given resource type if not detect first required reference
        reference_field = self._reference_fields.get(
            resource.get_resource_type(), self._get_required_reference(resource)
        )
        resource.__setattr__(reference_field, reference)

    @staticmethod
    def _get_required_reference(
        resource: Union[FHIRResourceModel, Type[FHIRAbstractModel]]
    ) -> str:
        fields = get_resource_fields(resource)
        required_fields = []
        for field in fields:
            if field.required:
                # todo check this further
                # add the reference to the required reference fields
                if field.type_ == ReferenceType:
                    required_fields.append(field.name)
            if field.name in ["patient", "subject"]:
                required_fields.append(field.name)
        # if there is only one reference field return it

        if len(required_fields) == 1:
            reference = required_fields[0]
        # iterate over the reference fields and return the best match patient -> subject
        else:
            reference = None
            for field in set(required_fields):
                if field == "patient":
                    reference = field
                    break
                elif field == "subject":
                    reference = field
                else:
                    raise ValueError(
                        f"No reference field found for {resource.get_resource_type()}"
                    )
        return reference

    def _store_generated_resource(
        self, resource: Union[FHIRResourceModel, FHIRAbstractModel], resource_type: str
    ):
        # todo improve this
        for store in self._dataset.resources:
            if store.resource_type == resource_type:
                store.resources.append(resource.dict(exclude_none=True))

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}(name={self.name}, resource_types={self._resource_types}, n={self.n},"
            f" generators={self.generators})>"
        )



from datetime import datetime
from enum import Enum
from typing import List, Union

import pendulum
from fhir.resources.resource import Resource
from pendulum.datetime import DateTime

from fhir_kindling.generators.base import BaseGenerator
from fhir_kindling.generators.resource_generator import ResourceGenerator
from fhir_kindling.serde.json import json_dict
from fhir_kindling.util.resources import check_resource_contains_field


class Frequencies(str, Enum):
    HOURLY = "hourly"
    DAILY = "daily"
    WEEKLY = "weekly"
    MONTHLY = "monthly"
    YEARLY = "yearly"


class TimeUnits(str, Enum):
    SECONDS = "s"
    MINUTES = "m"
    HOURS = "h"
    DAYS = "d"
    WEEKS = "w"
    MONTHS = "m"
    YEARS = "y"


class TimeSeriesGenerator(BaseGenerator):
    resource_generator: ResourceGenerator
    time_field: str
    start: DateTime
    end: Union[DateTime, None]
    freq: Union[Frequencies, str]
    n: Union[int, None]
    period: Union[int, None]
    period_unit: Union[TimeUnits, str, None]

    def __init__(
        self,
        resource_generator: ResourceGenerator,
        time_field: str,
        start: DateTime,
        end: Union[DateTime, None] = None,
        freq: Union[Frequencies, str] = Frequencies.DAILY,
        n: Union[int, None] = None,
        period: Union[int, None] = None,
        period_unit: Union[TimeUnits, str, None] = None,
    ) -> None:
        self.generator = resource_generator
        self.time_field = time_field
        self.n = n
        self._prev_time: Union[DateTime, None] = None
        self.generate_ids = True

        self._validate_args(freq, period, period_unit, start, end, n)

    def generate(
        self, generate_ids: bool = True, as_dict: bool = False
    ) -> Union[List[Resource], List[dict]]:
        self.generate_ids = generate_ids
        if self.n is None:
            return self._generate_by_end(as_dict=as_dict)
        else:
            return self._generate_by_n(as_dict=as_dict)

    def _generate_by_end(self, as_dict: bool) -> List[Resource]:
        resources = []
        current_time = self._get_next_time()
        while current_time < self.end:
            model = self._generate_resource(current_time, as_dict=as_dict)
            current_time = self._get_next_time()
            resources.append(model)

        return resources

    def _generate_by_n(self, as_dict: bool) -> List[Resource]:
        resources = []
        for _ in range(self.n):
            next_time = self._get_next_time()
            model = self._generate_resource(next_time, as_dict=as_dict)
            resources.append(model)
        return resources

    def _generate_resource(
        self, time: DateTime, as_dict: bool
    ) -> Union[Resource, dict]:
        resource = json_dict(self.generator.generate(generate_ids=self.generate_ids))
        resource[self.time_field] = time.isoformat()

        if as_dict:
            return resource

        model = self.generator.resource(**resource)
        return model

    def _get_next_time(self) -> datetime:
        """Get the next time in the series based on the frequency. Updates the internal state of the generator.

        Raises:
            ValueError: If the frequency is not valid

        Returns:
            _description
        """
        if not self._prev_time:
            self._prev_time = self.start
            return self.start

        next_time = None
        if self.freq == Frequencies.HOURLY:
            next_time = self._prev_time.add(hours=1)
        elif self.freq == Frequencies.DAILY:
            next_time = self._prev_time.add(days=1)
        elif self.freq == Frequencies.WEEKLY:
            next_time = self._prev_time.add(weeks=1)
        elif self.freq == Frequencies.MONTHLY:
            next_time = self._prev_time.add(months=1)
        elif self.freq == Frequencies.YEARLY:
            next_time = self._prev_time.add(years=1)
        else:
            raise ValueError(f"Invalid frequency: {self.freq}")
        #
        self._prev_time = next_time
        return next_time

    def _validate_args(self, freq, period, period_unit, start, end, n):
        if end is None and n is None:
            raise ValueError("Either end or n must be specified")
        if isinstance(freq, str):
            freq = Frequencies(freq)
        self.freq = freq
        # validate period input
        self._validate_period_input(period, period_unit)
        # validate time input
        self._validate_time_input(start, end)
        # check that the field is in the resource
        check_resource_contains_field(self.generator.resource, self.time_field)

    def _validate_period_input(self, period: int, period_unit: TimeUnits):
        if period is not None and period_unit is None:
            raise ValueError(
                "If period is specified, period_unit must also be specified"
            )
        if period is None and period_unit is not None:
            raise ValueError(
                "If period_unit is specified, period must also be specified"
            )
        if period is not None and period_unit is not None:
            if isinstance(period_unit, str):
                period_unit = TimeUnits(period_unit)
            self.period = period
            self.period_unit = period_unit

    def _validate_time_input(self, start: DateTime, end: DateTime):
        if isinstance(start, datetime):
            self.start = pendulum.instance(start)
        elif isinstance(start, DateTime):
            self.start = start
        else:
            raise ValueError(f"Invalid start datetime object: {type(start)}")

        if end:
            if isinstance(end, datetime):
                self.end = pendulum.instance(end)
            elif isinstance(end, DateTime):
                self.end = start
            else:
                raise ValueError(f"Invalid end datetime object: {type(end)}")
        else:
            self.end = None



from typing import Any, List, Optional, Union
from uuid import uuid4

from fhir.resources import get_fhir_model_class
from fhir.resources.fhirresourcemodel import FHIRResourceModel
from fhir.resources.resource import Resource
from pydantic import BaseModel

from fhir_kindling.generators.base import BaseGenerator
from fhir_kindling.generators.field_generator import FieldGenerator


class FieldValue(BaseModel):
    field: str
    value: Union[Any, List[Any]]
    list_field: bool = False


class GeneratorParameters(BaseModel):
    count: Optional[int] = None
    field_generators: Optional[List[FieldGenerator]] = None
    field_values: Optional[List[FieldValue]] = None

    # todo count and list in field values of same length


class ResourceGenerator(BaseGenerator):
    def __init__(
        self,
        resource: Union[str, Any],
        n: int = None,
        field_values: dict = None,
        disable_validation: bool = False,
        generator_parameters: GeneratorParameters = None,
    ):
        if not isinstance(resource, str):
            try:
                resource_type = resource.get_resource_type()
                self.resource = get_fhir_model_class(resource_type)
            except Exception:
                raise ValueError(
                    f"Resource must be a string or a FHIRResourceModel, got {type(resource)}"
                )
        else:
            self.resource = get_fhir_model_class(resource)
        self.params = generator_parameters
        self.field_values = field_values
        if self.field_values and not disable_validation:
            self._check_required_fields()
        self.disable_validation = disable_validation
        self.n = n
        self._value_iterators = {}
        # list to store the field names of all fields being generated
        self._generated_fields = set()
        self._validated = False

    def required_fields(self) -> List[str]:
        required_fields = []
        for field_name, field in self.resource.__fields__.items():
            if field.required:
                required_fields.append(field_name)
        return required_fields

    def fields(self):
        return self.resource.__fields__

    def generate(
        self,
        disable_validation: bool = False,
        generate_ids: bool = False,
        as_dict: bool = False,
    ) -> Union[Resource, List[Resource]]:
        self.disable_validation = disable_validation
        # if field values are given parse them into parameters
        if self.n and not self.params:
            self.params = GeneratorParameters(count=self.n)

        if self.field_values and self.n:
            self._parse_field_values()
        if not self.disable_validation:
            self._validate_params()
        resources = self._generate_resources(generate_ids=generate_ids, as_dict=as_dict)
        return resources

    def _generate_resources(self, generate_ids: bool, as_dict: bool = False):
        if self.params.count:
            resources = []
            for _ in range(self.params.count):
                resource = self._generate_resource(generate_ids, as_dict=as_dict)
                resources.append(resource)
            return resources
        else:
            return self._generate_resource(generate_ids, as_dict=as_dict)

    def _generate_resource(
        self, generate_id: bool, as_dict: bool = False
    ) -> Union[FHIRResourceModel, dict]:
        # construct a resource object to hold the generated fields
        resource = self.resource.construct()
        resource_data = {}
        # disable assignment validation
        resource.Config.validate_assignment = False
        if self.params.field_values:
            for field_value in self.params.field_values:
                # update resource with field value
                self._update_with_field_value(resource_data, field_value)

        if self.params.field_generators:
            for generator in self.params.field_generators:
                # update resource with generated field value
                self._update_with_field_generator(resource_data, generator)
        if generate_id:
            resource_data["id"] = str(uuid4())

        # validate resource when validation is enabled
        # if not self.disable_validation:
        #     resource = self.resource(**resource.dict(exclude_none=True))
        if as_dict:
            return resource_data
        resource = self.resource(**resource_data)
        return resource

    def _update_with_field_value(self, resource_data: dict, field_value: FieldValue):
        if isinstance(field_value.value, list) and not field_value.list_field:
            iterator = self._value_iterators.get(field_value.field)
            if not iterator:
                iterator = iter(field_value.value)
                self._value_iterators[field_value.field] = iterator
            resource_data[field_value.field] = next(iterator)
        else:
            resource_data[field_value.field] = field_value.value

    def _update_with_field_generator(
        self, resource_data: dict, field_generator: FieldGenerator
    ):
        value = field_generator.generate()
        resource_data[field_generator.field] = value

    def _check_required_fields(self):
        """
        Check if all required fields are being generated by the field generators or are given as field values

        Returns:

        """
        required_fields_set = set(self.required_fields())
        if not required_fields_set.issubset(self._generated_fields):
            raise ValueError(
                f"Required fields {required_fields_set - self._generated_fields} not generated, "
                f"generated fields: {self._generated_fields}"
            )

    def _parse_field_values(self):
        params = GeneratorParameters(count=self.n)

        # todo: add support for field_values as a list of dicts

        self.params = params

    def _validate_params(self):
        # skip validation if it has been previously validated
        if self._validated:
            return
        # validate field values
        if self.params.field_values:
            self._validate_field_values()

        # validate field generators
        if self.params.field_generators:
            self._validate_field_generators()

        # check that the required fields are being generated
        # self._check_required_fields()

        self._validated = True

    def _validate_field_values(self):
        field_values = self.params.field_values
        resource_count = self.params.count

        for field_value in field_values:
            # check for duplicates in generated fields
            if field_value.field in self._generated_fields:
                raise ValueError(
                    f"Field value {field_value.value} is already generated"
                )
            self._generated_fields.add(field_value.field)

            # check that the list length matches the resource count
            if isinstance(field_value.value, list):
                if (
                    len(field_value.value) != resource_count
                    and not field_value.list_field
                ):
                    raise ValueError(
                        f"Field value list length does not match resource count: {field_value.field}"
                        f"Items in field value list: {len(field_value.value)},"
                        f" resource count: {resource_count}"
                    )
            # todo check that the type of the item fits into the selected field

    def _validate_field_generators(self):
        generators = self.params.field_generators
        for resource_generator in generators:
            # check for duplicates in generated fields
            if resource_generator.field in self._generated_fields:
                raise ValueError(
                    f"Field generator {resource_generator.field} is already generated"
                )
            self._generated_fields.add(resource_generator.field)
            # todo validate generator values

    def __repr__(self):
        return f"<{self.__class__.__name__}(n={self.n}, resource={self.resource}>"



import os

import pytest
from dotenv import find_dotenv, load_dotenv

from fhir_kindling import FhirServer


@pytest.fixture(scope="session")
def api_url():
    """
    Base api url and env vars
    """
    load_dotenv(find_dotenv())

    return os.getenv("FHIR_API_URL", "http://test.fhir.org/r4")


@pytest.fixture(scope="session")
def fhir_server(api_url):
    server = FhirServer(
        api_address=api_url,
        client_id=os.getenv("CLIENT_ID"),
        client_secret=os.getenv("CLIENT_SECRET"),
        oidc_provider_url=os.getenv("OIDC_PROVIDER_URL"),
    )
    return server
